with System; use System;
with Ada.Unchecked_Deallocation;
with Ada.Unchecked_Conversion;
with Grt.Errors; use Grt.Errors;
with Grt.Processes; use Grt.Processes;
with Grt.Typedesc_Handling; use Grt.Typedesc_Handling;

package body Grt.Signals is
   function Get_Mode_Signal (Sig : Ghdl_Signal_Ptr) return Mode_Signal_Type
   is
   begin
      return Sig.Instance_Name.Sig_Mode;
   end Get_Mode_Signal;

   Sig_Instance_Name : Ghdl_Instance_Name_Acc;
   Last_Implicit_Signal : Ghdl_Signal_Ptr;
   Current_Resolv : Resolved_Signal_Acc := null;

   procedure Ghdl_Signal_Name (Name : Ghdl_Str_Len_Ptr;
                               Instance_Name : Ghdl_Instance_Name_Acc;
                               Mode : Mode_Signal_Type;
                               Kind : Kind_Signal_Type;
                               Type_Desc : Ghdl_Desc_Ptr)
   is
      Instance : Ghdl_Instance_Name_Acc;
   begin
      Instance := new Ghdl_Instance_Name_Type'
        (Kind => Ghdl_Name_Signal,
         Name => Name,
         Parent => Instance_Name,
         Brother => Instance_Name.Child,
         Sig_Mode => Mode,
         Sig_Kind => Kind,
         Sig_Index => Sig_Table.Last + 1,
         Sig_Type_Desc => Type_Desc);
      Instance_Name.Child := Instance;
      Sig_Instance_Name := Instance;
   end Ghdl_Signal_Name;

   function Create_Signal
     (Mode : Mode_Type;
      Init_Val : Value_Union;
      Type_Desc : Ghdl_Desc_Ptr;
      Resolv_Func : System.Address;
      Resolv_Inst : System.Address)
     return Ghdl_Signal_Ptr
   is
      Res : Ghdl_Signal_Ptr;
      Resolv : Resolved_Signal_Acc;
   begin
      Sig_Table.Increment_Last;

      if Current_Resolv = null then
         if Resolv_Func /= Null_Address then
            Resolv := new Resolved_Signal_Type'
              (Resolv_Func => Resolv_Func,
               Resolv_Inst => Resolv_Inst,
               Type_Desc => Type_Desc,
               First_Signal => Sig_Table.Last,
               Last_Signal => Sig_Table.Last);
         else
            Resolv := null;
         end if;
      else
         if Resolv_Func /= Null_Address then
            --  Only one resolution function is allowed!
            Internal_Error ("create_signal");
         end if;
         Resolv := Current_Resolv;
         if Current_Resolv.Last_Signal = Sig_Table.Last then
            Current_Resolv := null;
         end if;
      end if;

      Res := new Ghdl_Signal'(Value => Init_Val,
                              Driving_Value => Init_Val,
                              Last_Value => Init_Val,
                              Last_Event => Std_Time'First,
                              Last_Active => Std_Time'First,
                              Event => False,
                              Active => False,
                              Nbr_Drivers => 0,
                              Drivers => null,
                              Nbr_Ports => 0,
                              Ports => null,
                              Event_List => null,
                              Effective => null,
                              Mode => Mode,
                              Propag => Propag_None,
                              Type_Desc => Type_Desc,
                              Instance_Name => Sig_Instance_Name,
                              Resolv => Resolv,
                              Time => 0,
                              Guard_Func => null,
                              Guard_Instance => System.Null_Address);

      --  Put the signal in the table.

      Sig_Table.Table (Sig_Table.Last) := Res;

      return Res;
   end Create_Signal;

   procedure Ghdl_Signal_Create_Resolution (Type_Desc : Ghdl_Desc_Ptr;
                                            Resolv_Func : System.Address;
                                            Resolv_Inst : System.Address)
   is
   begin
      if Current_Resolv /= null then
         Internal_Error ("Ghdl_Signal_Create_Resolution");
      end if;
      Current_Resolv := new Resolved_Signal_Type'
        (Resolv_Func => Resolv_Func,
         Resolv_Inst => Resolv_Inst,
         Type_Desc => Type_Desc,
         First_Signal => Sig_Table.Last + 1,
         Last_Signal => Sig_Table.Last
         + Sig_Table_Index (Get_Nbr_Signals (Type_Desc)));
   end Ghdl_Signal_Create_Resolution;

   procedure Check_New_Source (Sig : Ghdl_Signal_Ptr) is
   begin
      if Sig.Nbr_Drivers + Sig.Nbr_Ports > 0 and then Sig.Resolv = null then
         Error_C ("several sources for unresolved signal: ");
         Error_C (Sig.Instance_Name);
         Error_E ("");
      end if;
   end Check_New_Source;

   procedure Ghdl_Process_Add_Driver (Sign : Ghdl_Signal_Ptr)
   is
      type Size_T is new Integer;

      function Malloc (Size : Size_T) return Driver_Arr_Ptr;
      pragma Import (C, Malloc);

      function Realloc (Ptr : Driver_Arr_Ptr; Size : Size_T)
        return Driver_Arr_Ptr;
      pragma Import (C, Realloc);

      function Size (N : Natural) return Size_T is
      begin
         return Size_T (N * Driver_Type'Size / System.Storage_Unit);
      end Size;

      Trans : Transaction_Acc;
      Id : Process_Id;
   begin
      Id := Get_Current_Process_Id;
      if Sign.Nbr_Drivers = 0 then
         Check_New_Source (Sign);
         Sign.Drivers := Malloc (Size (1));
         Sign.Nbr_Drivers := 1;
      else
         -- Do not create a driver twice.
         for I in 0 .. Sign.Nbr_Drivers - 1 loop
            if Sign.Drivers (I).Proc = Id then
               return;
            end if;
         end loop;
         Check_New_Source (Sign);
         Sign.Nbr_Drivers := Sign.Nbr_Drivers + 1;
         Sign.Drivers := Realloc (Sign.Drivers, Size (Sign.Nbr_Drivers));
      end if;
      Trans := new Transaction'(Kind => Trans_Value,
                                Time => 0,
                                Next => null,
                                Val => Sign.Value);
      Sign.Drivers (Sign.Nbr_Drivers - 1) :=
        (First_Trans => Trans,
         Last_Trans => Trans,
         Proc => Id);
   end Ghdl_Process_Add_Driver;

   procedure Append_Port (Targ : Ghdl_Signal_Ptr; Src : Ghdl_Signal_Ptr)
   is
      type Size_T is new Integer;

      function Malloc (Size : Size_T) return Signal_Arr_Ptr;
      pragma Import (C, Malloc);

      function Realloc (Ptr : Signal_Arr_Ptr; Size : Size_T)
        return Signal_Arr_Ptr;
      pragma Import (C, Realloc);

      function Size (N : Natural) return Size_T is
      begin
         return Size_T (N * Ghdl_Signal_Ptr'Size / System.Storage_Unit);
      end Size;
   begin
      if Targ.Nbr_Ports = 0 then
         Targ.Ports := Malloc (Size (1));
         Targ.Nbr_Ports := 1;
      else
         Targ.Nbr_Ports := Targ.Nbr_Ports + 1;
         Targ.Ports := Realloc (Targ.Ports, Size (Targ.Nbr_Ports));
      end if;
      Targ.Ports (Targ.Nbr_Ports - 1) := Src;
   end Append_Port;

   --  Add SRC to port list of TARG, but only if not already in this list.
   procedure Add_Port (Targ : Ghdl_Signal_Ptr; Src : Ghdl_Signal_Ptr)
   is
   begin
      for I in 0 .. Targ.Nbr_Ports - 1 loop
         if Targ.Ports (I) = Src then
            return;
         end if;
      end loop;
      Append_Port (Targ, Src);
   end Add_Port;

   procedure Ghdl_Signal_Add_Source (Targ : Ghdl_Signal_Ptr;
                                     Src : Ghdl_Signal_Ptr)
   is
   begin
      Check_New_Source (Targ);
      Append_Port (Targ, Src);
   end Ghdl_Signal_Add_Source;

   procedure Free is new Ada.Unchecked_Deallocation
     (Object => Transaction, Name => Transaction_Acc);

   function Value_Equal (Left, Right : Value_Union; Mode : Mode_Type)
     return Boolean
   is
   begin
      case Mode is
         when Mode_B2 =>
            return Left.B2 = Right.B2;
         when Mode_E8 =>
            return Left.E8 = Right.E8;
         when Mode_E32 =>
            return Left.E32 = Right.E32;
         when Mode_I32 =>
            return Left.I32 = Right.I32;
         when Mode_P64 =>
            return Left.P64 = Right.P64;
         when Mode_F64 =>
            return Left.F64 = Right.F64;
      end case;
   end Value_Equal;

   function Find_Driver (Sig : Ghdl_Signal_Ptr) return Natural
   is
      Id : Process_Id;
   begin
      if Sig.Drivers = null then
         Error ("assignment to a signal without any driver");
      end if;
      Id := Get_Current_Process_Id;
      for I in 0 .. Sig.Nbr_Drivers - 1 loop
         if Sig.Drivers (I).Proc = Id then
            return I;
         end if;
      end loop;
      Error ("assignment to a signal without a driver for the process");
   end Find_Driver;

   procedure Ghdl_Signal_Start_Assign (Sign : Ghdl_Signal_Ptr;
                                       Reject : Std_Time;
                                       Trans : Transaction_Acc;
                                       After : Std_Time)
   is
      Assign_Time : Std_Time;
      Drv : constant Natural := Find_Driver (Sign);
      Driver : Driver_Type renames Sign.Drivers (Drv);
   begin
      --  LRM93 8.4.1
      --  It is an error if the time expression in a waveform element
      --  evaluates to a negative value.
      if After < 0 then
         Error ("negative time expression in signal assignment");
      end if;
      Assign_Time := Current_Time + After;

      --  LRM93 8.4.1
      --  1. All old transactions that are projected to occur at or after the
      --     time at which the earliest new transaction is projected to occur
      --     are deleted from the projected output waveform.
      if Driver.Last_Trans.Time >= Assign_Time then
         declare
            --  LAST is the last transaction to keep.
            Last : Transaction_Acc;
            Next : Transaction_Acc;
         begin
            Last := Driver.First_Trans;
            --  Find the first transaction to be deleted.
            Next := Last.Next;
            while Next /= null and then Next.Time < Assign_Time loop
               Last := Next;
               Next := Next.Next;
            end loop;
            --  Delete old transactions.
            if Next /= null then
               --  Set the last transaction of the driver.
               Driver.Last_Trans := Last;
               --  Cut the chain.  This is not strickly necessary, since
               --  it will be overriden below, by appending TRANS to the
               --  driver.
               Last.Next := null;
               --  Free removed transactions.
               loop
                  Last := Next.Next;
                  Free (Next);
                  exit when Last = null;
                  Next := Last;
               end loop;
            end if;
         end;
      end if;

      --  2.  The new transaction are then appended to the projected output
      --      waveform in the order of their projected occurence.
      Trans.Time := Assign_Time;
      Driver.Last_Trans.Next := Trans;
      Driver.Last_Trans := Trans;

      --  If the initial delay is inertial delay according to the definitions
      --  of section 8.4, the projected output waveform is further modified
      --  as follows:
      --  1.  All of the new transactions are marked.
      --  2.  An old transaction is marked if the time at which it is projected
      --      to occur is less than the time at which the first new transaction
      --      is projected to occur minus the pulse rejection limit.
      --  3.  For each remaining unmarked, old transaction, the old transaction
      --      is marked if it immediatly precedes a marked transaction and its
      --      value component is the same as that of the marked transaction;
      --  4.  The transaction that determines the current value of the driver
      --      is marked.
      --  5.  All unmarked transactions (all of which are old transactions) are
      --      deleted from the projected output waveform.
      --
      --  GHDL: only transactions that are projected to occur at [T-R, T[
      --  can be deleted (R is the reject time, T is now + after time).
      if Reject > 0 then
         --  LRM93 8.4
         --  It is an error if the pulse rejection limit for any inertially
         --  delayed signal assignment statement is [...] or greater than the
         --  time expression associated with the first waveform element.
         if Reject > After then
            Error ("pulse rejection greater than first waveform delay");
         end if;

         declare
            Prev : Transaction_Acc;
            Next : Transaction_Acc;
         begin
            --  Find the first transaction after the project time less the
            --  rejection time.
            --  PREV will be the last old transaction which is projected to
            --  occur before T - R.
            Prev := Driver.First_Trans;
            loop
               Next := Prev.Next;
               exit when Next.Time >= Assign_Time - Reject;
               Prev := Next;
            end loop;

            --  Scan every transaction until TRANS.  If a transaction value is
            --  different from the TRANS value, then delete all previous
            --  transactions (from T - R to the currently scanned transaction),
            --  since they are not marked.
            while Next /= Trans loop
               if Next.Kind /= Trans.Kind
                 or else
                 (Trans.Kind = Trans_Value
                  and then not Value_Equal (Next.Val, Trans.Val, Sign.Mode))
               then
                  --  NEXT is different from TRANS.
                  --  Delete ]PREV;NEXT].
                  declare
                     D, N : Transaction_Acc;
                  begin
                     D := Prev.Next;
                     Next := Next.Next;
                     Prev.Next := Next;
                     loop
                        N := D.Next;
                        Free (D);
                        exit when N = Next;
                        D := N;
                     end loop;
                  end;
               else
                  Next := Next.Next;
               end if;
            end loop;
         end;
      elsif Reject /= 0 then
         --  LRM93 8.4
         --  It is an error if the pulse rejection limit for any inertially
         --  delayed signal assignment statement is either negative or [...].
         Error ("pulse rejection is negative");
      end if;

      --  Do some checks.
      if Driver.Last_Trans.Next /= null then
         Error ("ghdl_signal_start_assign internal_error");
      end if;
   end Ghdl_Signal_Start_Assign;

   procedure Ghdl_Signal_Next_Assign (Sign : Ghdl_Signal_Ptr;
                                      Val : Value_Union;
                                      After : Std_Time)
   is
      Driver : Driver_Type renames Sign.Drivers (Find_Driver (Sign));
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'(Kind => Trans_Value,
                                Time => Current_Time + After,
                                Next => null,
                                Val => Val);
      Driver.Last_Trans.Next := Trans;
      Driver.Last_Trans := Trans;
   end Ghdl_Signal_Next_Assign;

   procedure Ghdl_Signal_Simple_Assign_Error (Sign : Ghdl_Signal_Ptr)
   is
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'(Kind => Trans_Error,
                                Time => 0,
                                Next => null);
      Ghdl_Signal_Start_Assign (Sign, 0, Trans, 0);
   end Ghdl_Signal_Simple_Assign_Error;

   procedure Ghdl_Signal_Start_Assign_Error (Sign : Ghdl_Signal_Ptr;
                                             Rej : Std_Time;
                                             After : Std_Time)
   is
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'(Kind => Trans_Error,
                                Time => 0,
                                Next => null);
      Ghdl_Signal_Start_Assign (Sign, Rej, Trans, After);
   end Ghdl_Signal_Start_Assign_Error;

   procedure Ghdl_Signal_Start_Assign_Null (Sign : Ghdl_Signal_Ptr;
                                            Rej : Std_Time;
                                            After : Std_Time)
   is
      Trans : Transaction_Acc;
   begin
      if Sign.Instance_Name.Sig_Kind = Kind_Signal_No then
         Error ("null transaction for a non-guarded target");
      end if;
      Trans := new Transaction'(Kind => Trans_Null,
                                Time => 0,
                                Next => null);
      Ghdl_Signal_Start_Assign (Sign, Rej, Trans, After);
   end Ghdl_Signal_Start_Assign_Null;

   function Ghdl_Create_Signal_B2
     (Init_Val : Ghdl_B2;
      Type_Desc : Ghdl_Desc_Ptr;
      Resolv_Func : System.Address;
      Resolv_Inst : System.Address)
     return Ghdl_Signal_Ptr
   is
   begin
      return Create_Signal
        (Mode_B2, Value_Union'(Mode => Mode_B2, B2 => Init_Val), Type_Desc,
         Resolv_Func, Resolv_Inst);
   end Ghdl_Create_Signal_B2;

   procedure Ghdl_Signal_Simple_Assign_B2 (Sign : Ghdl_Signal_Ptr;
                                           Val : Ghdl_B2)
   is
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'
        (Kind => Trans_Value,
         Time => 0,
         Next => null,
         Val => Value_Union'(Mode => Mode_B2, B2 => Val));

      Ghdl_Signal_Start_Assign (Sign, 0, Trans, 0);
   end Ghdl_Signal_Simple_Assign_B2;

   procedure Ghdl_Signal_Start_Assign_B2 (Sign : Ghdl_Signal_Ptr;
                                          Rej : Std_Time;
                                          Val : Ghdl_B2;
                                          After : Std_Time)
   is
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'
        (Kind => Trans_Value,
         Time => 0,
         Next => null,
         Val => Value_Union'(Mode => Mode_B2, B2 => Val));
      Ghdl_Signal_Start_Assign (Sign, Rej, Trans, After);
   end Ghdl_Signal_Start_Assign_B2;

   procedure Ghdl_Signal_Next_Assign_B2 (Sign : Ghdl_Signal_Ptr;
                                         Val : Ghdl_B2;
                                         After : Std_Time)
   is
   begin
      Ghdl_Signal_Next_Assign
        (Sign, Value_Union'(Mode => Mode_B2, B2 => Val), After);
   end Ghdl_Signal_Next_Assign_B2;

   function Ghdl_Create_Signal_E8
     (Init_Val : Ghdl_E8;
      Type_Desc : Ghdl_Desc_Ptr;
      Resolv_Func : System.Address;
      Resolv_Inst : System.Address)
     return Ghdl_Signal_Ptr
   is
   begin
      return Create_Signal
        (Mode_E8, Value_Union'(Mode => Mode_E8, E8 => Init_Val), Type_Desc,
         Resolv_Func, Resolv_Inst);
   end Ghdl_Create_Signal_E8;

   procedure Ghdl_Signal_Simple_Assign_E8 (Sign : Ghdl_Signal_Ptr;
                                           Val : Ghdl_E8)
   is
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'
        (Kind => Trans_Value,
         Time => 0,
         Next => null,
         Val => Value_Union'(Mode => Mode_E8, E8 => Val));

      Ghdl_Signal_Start_Assign (Sign, 0, Trans, 0);
   end Ghdl_Signal_Simple_Assign_E8;

   procedure Ghdl_Signal_Start_Assign_E8 (Sign : Ghdl_Signal_Ptr;
                                          Rej : Std_Time;
                                          Val : Ghdl_E8;
                                          After : Std_Time)
   is
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'
        (Kind => Trans_Value,
         Time => 0,
         Next => null,
         Val => Value_Union'(Mode => Mode_E8, E8 => Val));
      Ghdl_Signal_Start_Assign (Sign, Rej, Trans, After);
   end Ghdl_Signal_Start_Assign_E8;

   procedure Ghdl_Signal_Next_Assign_E8 (Sign : Ghdl_Signal_Ptr;
                                         Val : Ghdl_E8;
                                         After : Std_Time)
   is
   begin
      Ghdl_Signal_Next_Assign
        (Sign, Value_Union'(Mode => Mode_E8, E8 => Val), After);
   end Ghdl_Signal_Next_Assign_E8;

   function Ghdl_Create_Signal_I32
     (Init_Val : Ghdl_I32;
      Type_Desc : Ghdl_Desc_Ptr;
      Resolv_Func : System.Address;
      Resolv_Inst : System.Address)
     return Ghdl_Signal_Ptr
   is
   begin
      return Create_Signal
        (Mode_I32, Value_Union'(Mode => Mode_I32, I32 => Init_Val), Type_Desc,
         Resolv_Func, Resolv_Inst);
   end Ghdl_Create_Signal_I32;

   procedure Ghdl_Signal_Simple_Assign_I32 (Sign : Ghdl_Signal_Ptr;
                                            Val : Ghdl_I32)
   is
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'
        (Kind => Trans_Value,
         Time => 0,
         Next => null,
         Val => Value_Union'(Mode => Mode_I32, I32 => Val));

      Ghdl_Signal_Start_Assign (Sign, 0, Trans, 0);
   end Ghdl_Signal_Simple_Assign_I32;

   procedure Ghdl_Signal_Start_Assign_I32 (Sign : Ghdl_Signal_Ptr;
                                           Rej : Std_Time;
                                           Val : Ghdl_I32;
                                           After : Std_Time)
   is
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'
        (Kind => Trans_Value,
         Time => 0,
         Next => null,
         Val => Value_Union'(Mode => Mode_I32, I32 => Val));
      Ghdl_Signal_Start_Assign (Sign, Rej, Trans, After);
   end Ghdl_Signal_Start_Assign_I32;

   procedure Ghdl_Signal_Next_Assign_I32 (Sign : Ghdl_Signal_Ptr;
                                          Val : Ghdl_I32;
                                          After : Std_Time)
   is
   begin
      Ghdl_Signal_Next_Assign
        (Sign, Value_Union'(Mode => Mode_I32, I32 => Val), After);
   end Ghdl_Signal_Next_Assign_I32;

   function Ghdl_Create_Signal_P64
     (Init_Val : Ghdl_P64;
      Type_Desc : Ghdl_Desc_Ptr;
      Resolv_Func : System.Address;
      Resolv_Inst : System.Address)
     return Ghdl_Signal_Ptr
   is
   begin
      return Create_Signal
        (Mode_P64, Value_Union'(Mode => Mode_P64, P64 => Init_Val), Type_Desc,
         Resolv_Func, Resolv_Inst);
   end Ghdl_Create_Signal_P64;

   procedure Ghdl_Signal_Simple_Assign_P64 (Sign : Ghdl_Signal_Ptr;
                                            Val : Ghdl_P64)
   is
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'
        (Kind => Trans_Value,
         Time => 0,
         Next => null,
         Val => Value_Union'(Mode => Mode_P64, P64 => Val));

      Ghdl_Signal_Start_Assign (Sign, 0, Trans, 0);
   end Ghdl_Signal_Simple_Assign_P64;

   procedure Ghdl_Signal_Start_Assign_P64 (Sign : Ghdl_Signal_Ptr;
                                           Rej : Std_Time;
                                           Val : Ghdl_P64;
                                           After : Std_Time)
   is
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'
        (Kind => Trans_Value,
         Time => 0,
         Next => null,
         Val => Value_Union'(Mode => Mode_P64, P64 => Val));
      Ghdl_Signal_Start_Assign (Sign, Rej, Trans, After);
   end Ghdl_Signal_Start_Assign_P64;

   procedure Ghdl_Signal_Next_Assign_P64 (Sign : Ghdl_Signal_Ptr;
                                          Val : Ghdl_P64;
                                          After : Std_Time)
   is
   begin
      Ghdl_Signal_Next_Assign
        (Sign, Value_Union'(Mode => Mode_P64, P64 => Val), After);
   end Ghdl_Signal_Next_Assign_P64;

   function Ghdl_Create_Signal_F64
     (Init_Val : Ghdl_F64;
      Type_Desc : Ghdl_Desc_Ptr;
      Resolv_Func : System.Address;
      Resolv_Inst : System.Address)
     return Ghdl_Signal_Ptr
   is
   begin
      return Create_Signal
        (Mode_F64, Value_Union'(Mode => Mode_F64, F64 => Init_Val), Type_Desc,
         Resolv_Func, Resolv_Inst);
   end Ghdl_Create_Signal_F64;

   procedure Ghdl_Signal_Simple_Assign_F64 (Sign : Ghdl_Signal_Ptr;
                                            Val : Ghdl_F64)
   is
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'
        (Kind => Trans_Value,
         Time => 0,
         Next => null,
         Val => Value_Union'(Mode => Mode_F64, F64 => Val));

      Ghdl_Signal_Start_Assign (Sign, 0, Trans, 0);
   end Ghdl_Signal_Simple_Assign_F64;

   procedure Ghdl_Signal_Start_Assign_F64 (Sign : Ghdl_Signal_Ptr;
                                           Rej : Std_Time;
                                           Val : Ghdl_F64;
                                           After : Std_Time)
   is
      Trans : Transaction_Acc;
   begin
      Trans := new Transaction'
        (Kind => Trans_Value,
         Time => 0,
         Next => null,
         Val => Value_Union'(Mode => Mode_F64, F64 => Val));
      Ghdl_Signal_Start_Assign (Sign, Rej, Trans, After);
   end Ghdl_Signal_Start_Assign_F64;

   procedure Ghdl_Signal_Next_Assign_F64 (Sign : Ghdl_Signal_Ptr;
                                          Val : Ghdl_F64;
                                          After : Std_Time)
   is
   begin
      Ghdl_Signal_Next_Assign
        (Sign, Value_Union'(Mode => Mode_F64, F64 => Val), After);
   end Ghdl_Signal_Next_Assign_F64;

   procedure Ghdl_Signal_Internal_Checks
   is
      Sig : Ghdl_Signal_Ptr;
   begin
      for I in Sig_Table.First .. Sig_Table.Last loop
         Sig := Sig_Table.Table (I);

         --  Check drivers.
         for J in 0 .. Sig.Nbr_Drivers - 1 loop
            declare
               Trans : Transaction_Acc;
            begin
               Trans := Sig.Drivers (J).First_Trans;
               while Trans.Next /= null loop
                  if Trans.Next.Time < Trans.Time then
                     Internal_Error ("ghdl_signal_internal_checks: "
                                     & "bad transaction order");
                  end if;
                  Trans := Trans.Next;
               end loop;
               if Trans /= Sig.Drivers (J).Last_Trans then
                  Internal_Error ("ghdl_signal_internal_checks: "
                                  & "last transaction mismatch");
               end if;
            end;
         end loop;
      end loop;
   end Ghdl_Signal_Internal_Checks;

   procedure Ghdl_Signal_Effective_Value (Targ : Ghdl_Signal_Ptr;
                                          Src : Ghdl_Signal_Ptr)
   is
   begin
      if Targ.Effective /= null then
         Error ("internal error: already effective value");
      end if;
      Targ.Effective := Src;
   end Ghdl_Signal_Effective_Value;

   function Ghdl_Create_Stable_Signal (Val : Std_Time) return Ghdl_Signal_Ptr
   is
      Res : Ghdl_Signal_Ptr;
   begin
      Sig_Instance_Name := new Ghdl_Instance_Name_Type'
        (Kind => Ghdl_Name_Signal,
         Name => null,
         Parent => null,
         Brother => null,
         Sig_Mode => Mode_Stable,
         Sig_Kind => Kind_Signal_No,
         Sig_Index => Sig_Table.Last + 1,
         Sig_Type_Desc => To_Ghdl_Desc_Ptr (Std_Standard_Boolean_BTD'Address));
      Res := Create_Signal
        (Mode_B2, Value_Union'(Mode => Mode_B2, B2 => True),
         To_Ghdl_Desc_Ptr (Std_Standard_Boolean_BTD'Address),
         Null_Address, Null_Address);
      Res.Time := Val;
      Last_Implicit_Signal := Res;
      Ghdl_Process_Add_Driver (Res);
      return Res;
   end Ghdl_Create_Stable_Signal;

   procedure Ghdl_Signal_Register_Stable_Prefix (Sig : Ghdl_Signal_Ptr)
   is
   begin
      Add_Port (Last_Implicit_Signal, Sig);
   end Ghdl_Signal_Register_Stable_Prefix;

   Guard_String : constant String := "guard";
   Guard_Name : constant Ghdl_Str_Len_Address_Type :=
     (Len => 5, Str => Guard_String'Address);
   function To_Ghdl_Str_Len_Ptr is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Ghdl_Str_Len_Ptr);

   function Ghdl_Signal_Create_Guard (This : System.Address;
                                      Proc : Guard_Func_Acc;
                                      Instance_Name : Ghdl_Instance_Name_Acc)
     return Ghdl_Signal_Ptr
   is
      Res : Ghdl_Signal_Ptr;
   begin
      Sig_Instance_Name := new Ghdl_Instance_Name_Type'
        (Kind => Ghdl_Name_Signal,
         Name => To_Ghdl_Str_Len_Ptr (Guard_Name'Address),
         Parent => Instance_Name,
         Brother => Instance_Name.Child,
         Sig_Mode => Mode_Guard,
         Sig_Kind => Kind_Signal_No,
         Sig_Index => Sig_Table.Last + 1,
         Sig_Type_Desc => To_Ghdl_Desc_Ptr (Std_Standard_Boolean_BTD'Address));
      Instance_Name.Child := Sig_Instance_Name;
      Res := Create_Signal
        (Mode_B2, Value_Union'(Mode => Mode_B2, B2 => Proc.all (This)),
         To_Ghdl_Desc_Ptr (Std_Standard_Boolean_BTD'Address),
         Null_Address, Null_Address);
      Res.Guard_Func := Proc;
      Res.Guard_Instance := This;
      Last_Implicit_Signal := Res;
      return Res;
   end Ghdl_Signal_Create_Guard;

   procedure Ghdl_Signal_Guard_Dependence (Sig : Ghdl_Signal_Ptr)
   is
   begin
      Add_Port (Last_Implicit_Signal, Sig);
   end Ghdl_Signal_Guard_Dependence;

   function Find_Next_Time return Std_Time
   is
      Res : Std_Time;
      Sig : Ghdl_Signal_Ptr;
      Trans : Transaction_Acc;
   begin
      Res := Std_Time'Last;

      for I in Sig_Table.First .. Sig_Table.Last loop
         Sig := Sig_Table.Table (I);
         for J in 0 .. Sig.Nbr_Drivers - 1 loop
            Trans := Sig.Drivers (J).First_Trans.Next;
            if Trans /= null then
               if Trans.Time < Res then
                  Res := Trans.Time;
               end if;
               if Res = Current_Time then
                  return Res;
               end if;
            end if;
         end loop;
      end loop;
      return Res;
   end Find_Next_Time;

   type Resolv_B2_Func_Acc is access
     function (Length : Natural;
               Vector : System.Address;
               Instance : System.Address)
     return Ghdl_B2;
   function Address_To_Resolv_B2_Func_Acc is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Resolv_B2_Func_Acc);

   type Resolv_E8_Func_Acc is access
     function (Length : Natural;
               Vector : System.Address;
               Instance : System.Address)
     return Ghdl_E8;
   function Address_To_Resolv_E8_Func_Acc is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Resolv_E8_Func_Acc);

   type Resolv_I32_Func_Acc is access
     function (Length : Natural;
               Vector : System.Address;
               Instance : System.Address)
     return Ghdl_I32;
   function Address_To_Resolv_I32_Func_Acc is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Resolv_I32_Func_Acc);

   function Get_Nbr_Non_Null_Source (Sig : Ghdl_Signal_Ptr) return Natural
   is
      Length : Natural;
   begin
      Length := Sig.Nbr_Ports;
      for I in 0 .. Sig.Nbr_Drivers - 1 loop
         case Sig.Drivers (I).First_Trans.Kind is
            when Trans_Value =>
               Length := Length + 1;
            when Trans_Null =>
               null;
            when Trans_Error =>
               Error ("range check error");
         end case;
      end loop;
      return Length;
   end Get_Nbr_Non_Null_Source;

   procedure Compute_Simple_Resolved_Signal (Sig : Ghdl_Signal_Ptr)
   is
      Length : Natural;
      Index : Natural;
   begin
      --  Compute the number of non-null transaction.
      Length := Get_Nbr_Non_Null_Source (Sig);

      --  if 0 and register, exit.
      if Length = 0
        and then Sig.Instance_Name.Sig_Kind = Kind_Signal_Register
      then
         return;
      end if;

      --  create a vector of value
      case Sig.Mode is
         when Mode_B2 =>
            declare
               Vector : Ghdl_B2_Array (0 .. Length - 1);
               Func : Resolv_B2_Func_Acc;
            begin
               for I in 0 .. Sig.Nbr_Ports - 1 loop
                  Vector (I) := Sig.Ports (I).Driving_Value.B2;
               end loop;
               Index := Sig.Nbr_Ports;
               for I in 0 .. Sig.Nbr_Drivers - 1 loop
                  if Sig.Drivers (I).First_Trans.Kind = Trans_Value then
                     Vector (Index) := Sig.Drivers (I).First_Trans.Val.B2;
                     Index := Index + 1;
                  end if;
               end loop;
               Func := Address_To_Resolv_B2_Func_Acc (Sig.Resolv.Resolv_Func);
               Sig.Driving_Value.B2 := Func.all
                 (Length, Vector'Address, Sig.Resolv.Resolv_Inst);
            end;
         when Mode_E8 =>
            declare
               Vector : Ghdl_E8_Array (0 .. Length - 1);
               Func : Resolv_E8_Func_Acc;
            begin
               for I in 0 .. Sig.Nbr_Ports - 1 loop
                  Vector (I) := Sig.Ports (I).Driving_Value.E8;
               end loop;
               Index := Sig.Nbr_Ports;
               for I in 0 .. Sig.Nbr_Drivers - 1 loop
                  if Sig.Drivers (I).First_Trans.Kind = Trans_Value then
                     Vector (Index) := Sig.Drivers (I).First_Trans.Val.E8;
                     Index := Index + 1;
                  end if;
               end loop;
               Func := Address_To_Resolv_E8_Func_Acc (Sig.Resolv.Resolv_Func);
               Sig.Driving_Value.E8 := Func.all
                 (Length, Vector'Address, Sig.Resolv.Resolv_Inst);
            end;
         when Mode_I32 =>
            declare
               Vector : Ghdl_I32_Array (0 .. Length - 1);
               Func : Resolv_I32_Func_Acc;
            begin
               for I in 0 .. Sig.Nbr_Ports - 1 loop
                  Vector (I) := Sig.Ports (I).Driving_Value.I32;
               end loop;
               Index := Sig.Nbr_Ports;
               for I in 0 .. Sig.Nbr_Drivers - 1 loop
                  if Sig.Drivers (I).First_Trans.Kind = Trans_Value then
                     Vector (Index) := Sig.Drivers (I).First_Trans.Val.I32;
                     Index := Index + 1;
                  end if;
               end loop;
               Func := Address_To_Resolv_I32_Func_Acc (Sig.Resolv.Resolv_Func);
               Sig.Driving_Value.I32 := Func.all
                 (Length, Vector'Address, Sig.Resolv.Resolv_Inst);
            end;
         when others =>
            Internal_Error ("compute_simple_resolved_signal: bad mode");
      end case;
   end Compute_Simple_Resolved_Signal;

   type Multiple_Resolver_Acc is access procedure
     (Result : System.Address;
      Length : Ghdl_Index_Type;
      Vals : System.Address;
      Instance : System.Address);

   function Address_To_Multiple_Resolv is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Multiple_Resolver_Acc);

   procedure Compute_Multiple_Resolved_Signal (Resolv : Resolved_Signal_Acc)
   is
      Sig : Ghdl_Signal_Ptr;
      S : Ghdl_Signal_Ptr;
      Vec_Constr : Any_Constr_Type;
      Res_Constr : Any_Constr_Type;
      Length : Natural;
   begin
      --  A signal among others...
      Sig := Sig_Table.Table (Resolv.First_Signal);

      --  Compute number of non-null sources, set all signals active
      Length := Get_Nbr_Non_Null_Source (Sig);

      --  if 0 and register, exit.
      if Length = 0
        and then Sig.Instance_Name.Sig_Kind = Kind_Signal_Register
      then
         for J in Resolv.First_Signal .. Resolv.Last_Signal loop
            for I in 0 .. Sig.Nbr_Drivers - 1 loop
               if Sig_Table.Table (J).Drivers (I).First_Trans.Kind
                 /= Trans_Null
               then
                  Error ("null-transaction required");
               end if;
            end loop;
         end loop;
         return;
      end if;

      --  Allocate the value vector
      Create_Writer (Vec_Constr, Resolv.Type_Desc, Length);

      --  Fill it
      for I in 0 .. Sig.Nbr_Ports - 1 loop
         for J in Resolv.First_Signal .. Resolv.Last_Signal loop
            S := Sig_Table.Table (J).Ports (I);
            Write_Value (Vec_Constr, S.Mode, S.Driving_Value);
         end loop;
      end loop;
      for I in 0 .. Sig.Nbr_Drivers - 1 loop
         if Sig.Drivers (I).First_Trans.Kind = Trans_Null then
            for J in Resolv.First_Signal + 1 .. Resolv.Last_Signal loop
               if Sig_Table.Table (J).Drivers (I).First_Trans.Kind
                 /= Trans_Null
               then
                  Error ("null-transaction required");
               end if;
            end loop;
         else
            for J in Resolv.First_Signal .. Resolv.Last_Signal loop
               S := Sig_Table.Table (J);
               Write_Value (Vec_Constr, S.Mode, S.Drivers (I).First_Trans.Val);
            end loop;
         end if;
      end loop;

      Create_Reader (Res_Constr, Resolv.Type_Desc);

      --  call the resolution function
      Address_To_Multiple_Resolv (Resolv.Resolv_Func).all
        (Get_Value_Addr (Res_Constr),
         Ghdl_Index_Type (Length),
         Get_Value_Addr (Vec_Constr),
         Resolv.Resolv_Inst);

      --  extract the result
      Release (Vec_Constr);

      for I in Resolv.First_Signal .. Resolv.Last_Signal loop
         S := Sig_Table.Table (I);
         Read_Value (Res_Constr, S.Mode, S.Driving_Value);
      end loop;

      Release (Res_Constr);
   end Compute_Multiple_Resolved_Signal;

   procedure Resume_Process_If_Event
     (Sig : Ghdl_Signal_Ptr; Proc : Process_Id)
   is
      El : Action_List_Acc;
   begin
      El := new Action_List'(Kind => Action_Process,
                             Proc => Proc,
                             Next => Sig.Event_List);
      Sig.Event_List := El;
   end Resume_Process_If_Event;

   --  Order of signals:
   --  To be computed: driving value or/and effective value
   --  To be considered: ports, signals, implicit signals, resolution,
   --                    conversion
   --

   procedure Add_Propagation (P : Propagation_Type) is
   begin
      Propagation.Increment_Last;
      Propagation.Table (Propagation.Last) := P;
   end Add_Propagation;

   --  Put SIG in PROPAGATION table until ORDER level.
   procedure Order_Signal (Sig : Ghdl_Signal_Ptr; Order : Propag_Order_Flag);

   --  Return TRUE is the effective value of SIG is the driving value of SIG.
   function Is_Eff_Drv (Sig : Ghdl_Signal_Ptr) return Boolean
   is
   begin
      case Get_Mode_Signal (Sig) is
         when Mode_Signal
           | Mode_Buffer =>
            return True;
         when Mode_Linkage
           | Mode_Out =>
            --  No effective value.
            return False;
         when Mode_Inout
           | Mode_In =>
            if Sig.Effective = null then
               if Sig.Nbr_Drivers > 0 or Sig.Nbr_Ports > 0 then
                  --  Only for inout.
                  return True;
               else
                  return False;
               end if;
            else
               return False;
            end if;
         when Mode_Stable
           | Mode_Guard =>
            return True;
      end case;
   end Is_Eff_Drv;

   procedure Order_Signal_List (Sig : Ghdl_Signal_Ptr;
                                Order : Propag_Order_Flag)
   is
   begin
      for I in 0 .. Sig.Nbr_Ports - 1 loop
         Order_Signal (Sig.Ports (I), Order);
      end loop;
   end Order_Signal_List;

   --  Put SIG in PROPAGATION table until ORDER level.
   procedure Order_Signal (Sig : Ghdl_Signal_Ptr; Order : Propag_Order_Flag)
   is
      Sig_Mode : constant Mode_Signal_Type := Get_Mode_Signal (Sig);
   begin
      if Sig = null then
         return;
      end if;

      --  First, the driving value.
      case Sig.Propag is
         when Propag_None =>
            null;
         when Propag_Being_Driving =>
            Internal_Error ("order_signal: being driving");
         when Propag_Being_Effective =>
            Internal_Error ("order_signal: being effective");
         when Propag_Driving =>
            null;
         when Propag_Done =>
            --  If sig was already handled, nothing to do!
            return;
      end case;

      if Sig.Propag = Propag_None and then Sig_Mode in Mode_Signal_User then
         if Sig.Nbr_Drivers = 0 and Sig.Nbr_Ports = 0 then
            --  No source.
            Sig.Propag := Propag_Driving;
         elsif Sig.Resolv = null then
            --  Not resolved (so at most one source).
            if Sig.Nbr_Drivers = 1 then
               --  Not resolved, 1 source : a driver.
               if Is_Eff_Drv (Sig) then
                  Add_Propagation ((Kind => Eff_One_Driver, Sig => Sig));
                  Sig.Propag := Propag_Done;
               else
                  Add_Propagation ((Kind => Drv_One_Driver, Sig => Sig));
                  Sig.Propag := Propag_Driving;
               end if;
            else
               Sig.Propag := Propag_Being_Driving;
               --  not resolved, 1 source : Source is a port.
               Order_Signal (Sig.Ports (0), Propag_Driving);
               if Is_Eff_Drv (Sig) then
                  Add_Propagation ((Kind => Eff_One_Port, Sig => Sig));
                  Sig.Propag := Propag_Done;
               else
                  Add_Propagation ((Kind => Drv_One_Port, Sig => Sig));
                  Sig.Propag := Propag_Driving;
               end if;
            end if;
         else
            --  Resolved signal.
            declare
               Resolv : Resolved_Signal_Acc;
               S : Ghdl_Signal_Ptr;
            begin
               --  Compute driving value of brothers.
               Resolv := Sig.Resolv;
               for I in Resolv.First_Signal .. Resolv.Last_Signal loop
                  S := Sig_Table.Table (I);
                  if S.Propag /= Propag_None then
                     Internal_Error ("order_signal(1)");
                  end if;
                  S.Propag := Propag_Being_Driving;
               end loop;
               for I in Resolv.First_Signal .. Resolv.Last_Signal loop
                  S := Sig_Table.Table (I);
                  --  Compute driving value of the sources.
                  for J in 0 .. S.Nbr_Ports - 1 loop
                     Order_Signal (S.Ports (J), Propag_Driving);
                  end loop;
               end loop;
               for I in Resolv.First_Signal .. Resolv.Last_Signal loop
                  S := Sig_Table.Table (I);
                  S.Propag := Propag_Driving;
               end loop;

               if Is_Eff_Drv (Sig) then
                  if Resolv.First_Signal = Resolv.Last_Signal then
                     Add_Propagation ((Kind => Eff_One_Resolved, Sig => Sig));
                  else
                     Add_Propagation ((Kind => Eff_Multiple,
                                       Resolv => Resolv));
                  end if;
               else
                  if Resolv.First_Signal = Resolv.Last_Signal then
                     Add_Propagation ((Kind => Drv_One_Resolved, Sig => Sig));
                  else
                     Add_Propagation ((Kind => Drv_Multiple,
                                       Resolv => Resolv));
                  end if;
               end if;
            end;
         end if;
      elsif Sig.Propag = Propag_None
        and then Sig_Mode in Mode_Signal_Implicit
      then
         case Mode_Signal_Implicit (Sig_Mode) is
            when Mode_Guard =>
               Sig.Propag := Propag_Being_Driving;
               Order_Signal_List (Sig, Propag_Done);
               Sig.Propag := Propag_Done;
               Add_Propagation ((Kind => Imp_Guard, Sig => Sig));
               return;
            when Mode_Stable =>
               Sig.Propag := Propag_Being_Driving;
               Order_Signal_List (Sig, Propag_Done);
               Sig.Propag := Propag_Done;
               Add_Propagation ((Kind => Imp_Stable, Sig => Sig));
               return;
         end case;
      end if;

      -- Effective value.
      if Order = Propag_Driving then
         --  Will be done later.
         return;
      end if;

      case Sig_Mode is
         when Mode_Signal
           | Mode_Buffer =>
            --  Effective value is driving value.
            Sig.Propag := Propag_Done;
         when Mode_Linkage
           | Mode_Out =>
            --  No effective value.
            Sig.Propag := Propag_Done;
         when Mode_Inout
           | Mode_In =>
            if Sig.Effective = null then
               --  Effective value is driving value or initial value.
               null;
            else
               Sig.Propag := Propag_Being_Effective;
               Order_Signal (Sig.Effective, Propag_Done);
               Add_Propagation ((Kind => Eff_Actual, Sig => Sig));
               Sig.Propag := Propag_Done;
            end if;
         when Mode_Stable
           | Mode_Guard =>
            null;
      end case;
   end Order_Signal;

   --  Check every scalar subelement of a resolved signal has a driver
   --  in the same process.
   procedure Check_Resolved_Driver (Resolv : Resolved_Signal_Acc)
   is
      First_Sig : Ghdl_Signal_Ptr;
      Nbr : Natural;
   begin
      First_Sig := Sig_Table.Table (Resolv.First_Signal);
      Nbr := First_Sig.Nbr_Drivers;
      for I in Resolv.First_Signal + 1 .. Resolv.Last_Signal loop
         if Sig_Table.Table (I).Nbr_Drivers /= Nbr then
            --  FIXME: provide more information (signal name, process name).
            Error ("missing drivers for subelement of a resolved signal");
         end if;
      end loop;
   end Check_Resolved_Driver;

   procedure Order_All_Signals
   is
      Sig : Ghdl_Signal_Ptr;
   begin
      for I in Sig_Table.First .. Sig_Table.Last loop
         Sig := Sig_Table.Table (I);

         --  Check every scalar subelement of a resolved signal has a driver
         --  in the same process.
         if Sig.Resolv /= null
           and then Sig.Resolv.First_Signal = I
           and then Sig.Resolv.Last_Signal > I
         then
            Check_Resolved_Driver (Sig.Resolv);
         end if;

         Order_Signal (Sig, Propag_Driving);
      end loop;
      for I in Sig_Table.First .. Sig_Table.Last loop
         Order_Signal (Sig_Table.Table (I), Propag_Done);
      end loop;
   end Order_All_Signals;

   procedure Set_Guard_Activity (Sig : Ghdl_Signal_Ptr) is
   begin
      for I in 0 .. Sig.Nbr_Ports - 1 loop
         if Sig.Ports (I).Active then
            Sig.Active := True;
            return;
         end if;
      end loop;
   end Set_Guard_Activity;

   procedure Set_Stable_Activity (Sig : Ghdl_Signal_Ptr) is
   begin
      for I in 0 .. Sig.Nbr_Ports - 1 loop
         if Sig.Ports (I).Event then
            Sig.Active := True;
            return;
         end if;
      end loop;
   end Set_Stable_Activity;

   procedure Set_Resolved_Activity (Sig : Ghdl_Signal_Ptr)
   is
      Trans : Transaction_Acc;
   begin
      for J in 0 .. Sig.Nbr_Drivers - 1 loop
         Trans := Sig.Drivers (J).First_Trans.Next;
         if Trans /= null and then Trans.Time = Current_Time then
            Free (Sig.Drivers (J).First_Trans);
            Sig.Drivers (J).First_Trans := Trans;
            Sig.Active := True;
         end if;
      end loop;
      for J in 0 .. Sig.Nbr_Ports - 1 loop
         if Sig.Ports (J).Active then
            Sig.Active := True;
            exit;
         end if;
      end loop;
   end Set_Resolved_Activity;

   procedure Set_Effective_Value (Sig : Ghdl_Signal_Ptr; Val : Value_Union)
   is
      El : Action_List_Acc;
   begin
      if not Value_Equal (Sig.Value, Val, Sig.Mode) then
         Sig.Last_Value := Sig.Value;
         Sig.Value := Val;
         Sig.Event := True;
         Sig.Last_Event := Current_Time;

         El := Sig.Event_List;
         while El /= null loop
            case El.Kind is
               when Action_Process =>
                  Resume_Process (El.Proc);
               when Action_Signal =>
                  Internal_Error ("set_effective_value");
            end case;
            El := El.Next;
         end loop;
      end if;
   end Set_Effective_Value;

   procedure Update_Signals
   is
      Sig : Ghdl_Signal_Ptr;
      Trans : Transaction_Acc;
   begin
      --  LRM93 12.6.2
      --  1) Reset active flag.
      for I in Sig_Table.First .. Sig_Table.Last loop
         Sig := Sig_Table.Table (I);
         Sig.Active := False;
         Sig.Event := False;
      end loop;

      for I in Propagation.First .. Propagation.Last loop
         --  First: the driving value.
         case Propagation.Table (I).Kind is
            when Drv_One_Driver
              | Eff_One_Driver =>
               Sig := Propagation.Table (I).Sig;
               Trans := Sig.Drivers (0).First_Trans.Next;
               if Trans /= null and then Trans.Time = Current_Time then
                  Sig.Active := True;
                  Free (Sig.Drivers (0).First_Trans);
                  Sig.Drivers (0).First_Trans := Trans;
                  case Trans.Kind is
                     when Trans_Value =>
                        Sig.Driving_Value := Trans.Val;
                     when Trans_Null =>
                        Error ("null transaction");
                     when Trans_Error =>
                        Error ("range check error on signal");
                  end case;
               end if;
            when Drv_One_Resolved
              | Eff_One_Resolved =>
               Sig := Propagation.Table (I).Sig;
               Set_Resolved_Activity (Sig);
               if Sig.Active then
                  Compute_Simple_Resolved_Signal (Propagation.Table (I).Sig);
               end if;
            when Drv_One_Port
              | Eff_One_Port =>
               Sig := Propagation.Table (I).Sig;
               if Sig.Ports (0).Active then
                  Sig.Active := True;
                  Sig.Driving_Value := Sig.Ports (0).Driving_Value;
               end if;
            when Eff_Actual =>
               Sig := Propagation.Table (I).Sig;
               --  Note: the signal may have drivers (inout ports).
               Sig.Active := Sig.Active or Sig.Effective.Active;
            when Drv_Multiple
              | Eff_Multiple =>
               declare
                  Active : Boolean := False;
                  Resolv : Resolved_Signal_Acc;
               begin
                  Resolv := Propagation.Table (I).Resolv;
                  for I in Resolv.First_Signal .. Resolv.Last_Signal loop
                     Sig := Sig_Table.Table (I);
                     Set_Resolved_Activity (Sig);
                     Active := Active or Sig.Active;
                  end loop;
                  if Active then
                     Compute_Multiple_Resolved_Signal (Resolv);
                  end if;
               end;
            when Imp_Guard
              | Imp_Stable =>
               null;
            when others =>
               Internal_Error ("update signals");
         end case;

         --  Second: the effective value.
         case Propagation.Table (I).Kind is
            when Drv_One_Driver
              | Drv_One_Port
              | Drv_One_Resolved
              | Drv_Multiple =>
               null;
            when Eff_One_Driver
              | Eff_One_Port
              | Eff_One_Resolved =>
               Sig := Propagation.Table (I).Sig;
               if Sig.Active then
                  Set_Effective_Value (Sig, Sig.Driving_Value);
               end if;
            when Eff_Multiple =>
               declare
                  Resolv : Resolved_Signal_Acc;
               begin
                  Resolv := Propagation.Table (I).Resolv;
                  if Sig_Table.Table (Resolv.First_Signal).Active then
                     --  If one signal is active, all are active.
                     for I in Resolv.First_Signal .. Resolv.Last_Signal loop
                        Sig := Sig_Table.Table (I);
                        Set_Effective_Value (Sig, Sig.Driving_Value);
                     end loop;
                  end if;
               end;
            when Eff_Actual =>
               Sig := Propagation.Table (I).Sig;
               if Sig.Active then
                  Set_Effective_Value (Sig, Sig.Effective.Value);
               end if;
            when Imp_Guard =>
               --  Guard signal is active iff one of its dependence is active.
               Sig := Propagation.Table (I).Sig;
               Set_Guard_Activity (Sig);
               if Sig.Active then
                  Sig.Driving_Value.B2 :=
                    Sig.Guard_Func.all (Sig.Guard_Instance);
                  Set_Effective_Value (Sig, Sig.Driving_Value);
               end if;
            when Imp_Stable =>
               Sig := Propagation.Table (I).Sig;
               Set_Stable_Activity (Sig);
               if Sig.Active then
                  Sig.Driving_Value :=
                    Value_Union'(Mode => Mode_B2, B2 => False);
                  Sig.Last_Active := Current_Time;
                  --  Set driver.
                  Trans := new Transaction'
                    (Kind => Trans_Value,
                     Time => Current_Time + Sig.Time,
                     Next => null,
                     Val => Value_Union'(Mode => Mode_B2, B2 => True));
                  if Sig.Drivers (0).First_Trans.Next /= null then
                     Free (Sig.Drivers (0).First_Trans.Next);
                  end if;
                  Sig.Drivers (0).First_Trans.Next := Trans;
                  Sig.Drivers (0).Last_Trans := Trans;
                  Set_Effective_Value (Sig, Sig.Driving_Value);
               else
                  Trans := Sig.Drivers (0).First_Trans.Next;
                  if Trans /= null and then Trans.Time = Current_Time then
                     Sig.Active := True;
                     Free (Sig.Drivers (0).First_Trans);
                     Sig.Drivers (0).First_Trans := Trans;
                     Sig.Driving_Value := Trans.Val;
                     Set_Effective_Value (Sig, Sig.Driving_Value);
                  end if;
               end if;
            when others =>
               Internal_Error ("update_signals(2)");
         end case;
      end loop;
   end Update_Signals;

   procedure Init_Signals
   is
      Sig : Ghdl_Signal_Ptr;
   begin
      for I in Propagation.First .. Propagation.Last loop
         --  First: the driving value.
         case Propagation.Table (I).Kind is
            when Drv_One_Driver
              | Eff_One_Driver =>
               --  Nothing to do: drivers were already created.
               null;
            when Drv_One_Resolved
              | Eff_One_Resolved =>
               --  Execute the resolution function.
               Compute_Simple_Resolved_Signal (Propagation.Table (I).Sig);
            when Drv_One_Port
              | Eff_One_Port =>
               --  Copy value.
               Sig := Propagation.Table (I).Sig;
               Sig.Driving_Value := Sig.Ports (0).Driving_Value;
            when Eff_Actual =>
               null;
            when Drv_Multiple
              | Eff_Multiple =>
               declare
                  Resolv : Resolved_Signal_Acc;
               begin
                  Resolv := Propagation.Table (I).Resolv;
                  Compute_Multiple_Resolved_Signal (Resolv);
               end;
            when Imp_Guard
              | Imp_Stable =>
               null;
            when others =>
               Internal_Error ("init_signals");
         end case;

         --  Second: the effective value.
         case Propagation.Table (I).Kind is
            when Drv_One_Driver
              | Drv_One_Port
              | Drv_One_Resolved
              | Drv_Multiple =>
               null;
            when Eff_One_Driver
              | Eff_One_Port
              | Eff_One_Resolved =>
               Sig := Propagation.Table (I).Sig;
               Sig.Value := Sig.Driving_Value;
            when Eff_Multiple =>
               declare
                  Resolv : Resolved_Signal_Acc;
               begin
                  Resolv := Propagation.Table (I).Resolv;
                  for I in Resolv.First_Signal .. Resolv.Last_Signal loop
                     Sig := Sig_Table.Table (I);
                     Sig.Value := Sig.Driving_Value;
                  end loop;
               end;
            when Eff_Actual =>
               Sig := Propagation.Table (I).Sig;
               Sig.Value := Sig.Effective.Value;
            when Imp_Guard =>
               --  Guard signal is active iff one of its dependence is active.
               Sig := Propagation.Table (I).Sig;
               Sig.Driving_Value.B2 :=
                 Sig.Guard_Func.all (Sig.Guard_Instance);
               Sig.Value := Sig.Driving_Value;
            when Imp_Stable =>
               --  Already initialized during creation.
               null;
            when others =>
               Internal_Error ("init_signals(2)");
         end case;
      end loop;
   end Init_Signals;

end Grt.Signals;
