with Interfaces; use Interfaces;
with System;
with GNAT.Table;
with Grt.Types; use Grt.Types;
with Grt.Hierarchy; use Grt.Hierarchy;
with Grt.Typedesc; use Grt.Typedesc;

package Grt.Signals is
   --  Kind of transaction.
   type Transaction_Kind is
     (
      --  Normal transaction, with a value.
      Trans_Value,
      --  Null transaction.
      Trans_Null,
      --  Like a normal transaction, but without a value due to check error.
      Trans_Error);

   type Transaction;
   type Transaction_Acc is access Transaction;
   type Transaction (Kind : Transaction_Kind) is record
      Time : Std_Time;
      Next : Transaction_Acc;
      case Kind is
         when Trans_Value =>
            Val : Value_Union;
         when Trans_Null =>
            null;
         when Trans_Error =>
            --  FIXME: should have a location field, to be able to display
            --  a message.
            null;
      end case;
   end record;

   --  A driver is bound to a process (PROC) and contains a list of
   --  transactions.
   type Driver_Type is record
      First_Trans : Transaction_Acc;
      Last_Trans : Transaction_Acc;
      Proc : Process_Id;
   end record;

   type Driver_Fat_Array is array (Natural) of Driver_Type;
   type Driver_Arr_Ptr is access Driver_Fat_Array;

   --  Function access type used to evaluate the guard expression.
   type Guard_Func_Acc is access function (This : System.Address)
     return Ghdl_B2;

   --  Simply linked list of processes to be resumed in case of events.

   type Ghdl_Signal;
   type Ghdl_Signal_Ptr is access Ghdl_Signal;

   type Signal_Fat_Array is array (Natural) of Ghdl_Signal_Ptr;
   type Signal_Arr_Ptr is access Signal_Fat_Array;

   type Action_List;
   type Action_List_Acc is access Action_List;
   type Action_Kind is (Action_Signal, Action_Process);
   type Action_List (Kind : Action_Kind) is record
      Next : Action_List_Acc;
      case Kind is
         when Action_Signal =>
            Sig : Ghdl_Signal_Ptr;
         when Action_Process =>
            Proc : Process_Id;
      end case;
   end record;

   --  How to compute resolved signal.
   type Resolved_Signal_Type is record
      Resolv_Func : System.Address;
      Resolv_Inst : System.Address;
      Type_Desc : Ghdl_Desc_Ptr;
      First_Signal : Sig_Table_Index;
      Last_Signal : Sig_Table_Index;
   end record;

   type Resolved_Signal_Acc is access Resolved_Signal_Type;

   --  Used to order the signals for the propagation of signals values.
   type Propag_Order_Flag is
     (
      --  The signal was not yet ordered.
      Propag_None,
      --  The signal is being ordered for driving value.
      --  This stage is used to catch loop (which can not occur).
      Propag_Being_Driving,
      --  The signal has been ordered for driving value.
      Propag_Driving,
      --  The signal is being ordered for effective value.
      Propag_Being_Effective,
      Propag_Done);

   type Ghdl_Signal is record
      --  Fields known by ghdl.
      Value : Value_Union;
      Driving_Value : Value_Union;
      Last_Value : Value_Union;
      Last_Event : Std_Time;
      Last_Active : Std_Time;
      Event : Boolean;
      Active : Boolean;

      --  Internal fields.
      --  Values mode of this signal.
      Mode : Mode_Type;

      --  Status of the ordering.
      Propag : Propag_Order_Flag;

      --  List of process to resume when there is an event on
      --  this signal.
      Event_List : Action_List_Acc;

      --  Signal which defines the effective value of this signal, if any.
      Effective : Ghdl_Signal_Ptr;

      --  Type description for this signal.
      Type_Desc : Ghdl_Desc_Ptr;

      --  Path of the signal (with its name) in the design hierarchy.
      Instance_Name : Ghdl_Instance_Name_Acc;

      --  For user signals: the sources of a signals are drivers
      --  and connected ports.
      --  For implicit signals: PORTS is used as dependence list.
      Nbr_Drivers : Natural;
      Drivers : Driver_Arr_Ptr;
      Nbr_Ports : Natural;
      Ports : Signal_Arr_Ptr;

      --  Null if not resolved.
      Resolv : Resolved_Signal_Acc;

      --  Time parameter for 'stable, 'quiet or 'delay
      Time : Std_Time;

      --  Guard function and instance used to compute the guard expression.
      Guard_Func : Guard_Func_Acc;
      Guard_Instance : System.Address;
   end record;

   --  Each simple signal declared can be accessed by SIG_TABLE.
   package Sig_Table is new GNAT.Table
     (Table_Component_Type => Ghdl_Signal_Ptr,
      Table_Index_Type => Sig_Table_Index,
      Table_Low_Bound => 0,
      Table_Initial => 128,
      Table_Increment => 100);

   --  Return the next time at which a driver becomes active.
   function Find_Next_Time return Std_Time;

   --  Elementary propagation computation.
   --  See LRM 12.6.2 and 12.6.3
   type Propagation_Kind_Type is
     (
      --  How to compute driving value:
      --  Default value.
      Drv_Error,

      --  One source, a driver and not resolved:
      --  the driving value is the driver.
      Drv_One_Driver,

      --  Same as previous, but the effective value is the driving value.
      Eff_One_Driver,

      --  One source, a port and not resolved:
      --  the driving value is the driving value of the port.
      --  Dependence.
      Drv_One_Port,

      --  Same as previous, but the effective value is the driving value.
      Eff_One_Port,

      --  Several sources or resolved:
      --  signal is not compite.
      Drv_One_Resolved,
      Eff_One_Resolved,

      --  Use the resolution function, signal is composite.
      Drv_Multiple,

      --  Same as previous, but the effective value is the previous value.
      Eff_Multiple,

      --  The effective value is the actual associated.
      Eff_Actual,

      --  Implicit guard signal.
      --  Its value must be evaluated after the effective value of its
      --  dependences.
      Imp_Guard,

      --  Implicit stable.
      --  Its value must be evaluated after the effective value of its
      --  dependences.
      Imp_Stable,

      --  Implicit quiet.
      --  Its value must be evaluated after the driving value of its
      --  dependences.
      Imp_Quiet
      );

   type Propagation_Type (Kind : Propagation_Kind_Type := Drv_Error) is record
      case Kind is
         when Drv_One_Driver
           | Eff_One_Driver
           | Drv_One_Port
           | Eff_One_Port
           | Imp_Guard
           | Imp_Stable
           | Eff_Actual
           | Eff_One_Resolved
           | Drv_One_Resolved =>
            Sig : Ghdl_Signal_Ptr;
         when Drv_Multiple
           | Eff_Multiple =>
            Resolv : Resolved_Signal_Acc;
         when others =>
            null;
      end case;
   end record;

   package Propagation is new GNAT.Table
     (Table_Component_Type => Propagation_Type,
      Table_Index_Type => Natural,
      Table_Low_Bound => 1,
      Table_Initial => 128,
      Table_Increment => 100);

   --  Compute propagation order of signals.
   procedure Order_All_Signals;

   --  Initialize all signals.
   procedure Init_Signals;

   --  Update signals.
   procedure Update_Signals;

   --  Add PROC in the list of processes to be resumed in case of event on
   --  SIG.
   procedure Resume_Process_If_Event
     (Sig : Ghdl_Signal_Ptr; Proc : Process_Id);

   --  Set the name of the next signal(s) to be registered.
   procedure Ghdl_Signal_Name (Name : Ghdl_Str_Len_Ptr;
                               Instance_Name : Ghdl_Instance_Name_Acc;
                               Mode : Mode_Signal_Type;
                               Kind : Kind_Signal_Type;
                               Type_Desc : Ghdl_Desc_Ptr);

   --  Performs some internal checks on signals (transaction order).
   --  Internal_error is called in case of error.
   procedure Ghdl_Signal_Internal_Checks;

   --  Subprograms to be called by generated code.
   procedure Ghdl_Signal_Simple_Assign_Error (Sign : Ghdl_Signal_Ptr);
   procedure Ghdl_Signal_Start_Assign_Error (Sign : Ghdl_Signal_Ptr;
                                             Rej : Std_Time;
                                             After : Std_Time);

   procedure Ghdl_Signal_Start_Assign_Null (Sign : Ghdl_Signal_Ptr;
                                            Rej : Std_Time;
                                            After : Std_Time);

   function Ghdl_Create_Signal_B2
     (Init_Val : Ghdl_B2;
      Type_Desc : Ghdl_Desc_Ptr;
      Resolv_Func : System.Address;
      Resolv_Inst : System.Address)
     return Ghdl_Signal_Ptr;
   procedure Ghdl_Signal_Simple_Assign_B2 (Sign : Ghdl_Signal_Ptr;
                                           Val : Ghdl_B2);
   procedure Ghdl_Signal_Start_Assign_B2 (Sign : Ghdl_Signal_Ptr;
                                          Rej : Std_Time;
                                          Val : Ghdl_B2;
                                          After : Std_Time);
   procedure Ghdl_Signal_Next_Assign_B2 (Sign : Ghdl_Signal_Ptr;
                                         Val : Ghdl_B2;
                                         After : Std_Time);

   function Ghdl_Create_Signal_E8
     (Init_Val : Ghdl_E8;
      Type_Desc : Ghdl_Desc_Ptr;
      Resolv_Func : System.Address;
      Resolv_Inst : System.Address)
     return Ghdl_Signal_Ptr;
   procedure Ghdl_Signal_Simple_Assign_E8 (Sign : Ghdl_Signal_Ptr;
                                           Val : Ghdl_E8);
   procedure Ghdl_Signal_Start_Assign_E8 (Sign : Ghdl_Signal_Ptr;
                                          Rej : Std_Time;
                                          Val : Ghdl_E8;
                                          After : Std_Time);
   procedure Ghdl_Signal_Next_Assign_E8 (Sign : Ghdl_Signal_Ptr;
                                         Val : Ghdl_E8;
                                         After : Std_Time);

   function Ghdl_Create_Signal_I32
     (Init_Val : Ghdl_I32;
      Type_Desc : Ghdl_Desc_Ptr;
      Resolv_Func : System.Address;
      Resolv_Inst : System.Address)
     return Ghdl_Signal_Ptr;
   procedure Ghdl_Signal_Simple_Assign_I32 (Sign : Ghdl_Signal_Ptr;
                                            Val : Ghdl_I32);
   procedure Ghdl_Signal_Start_Assign_I32 (Sign : Ghdl_Signal_Ptr;
                                           Rej : Std_Time;
                                           Val : Ghdl_I32;
                                           After : Std_Time);
   procedure Ghdl_Signal_Next_Assign_I32 (Sign : Ghdl_Signal_Ptr;
                                          Val : Ghdl_I32;
                                          After : Std_Time);

   function Ghdl_Create_Signal_P64
     (Init_Val : Ghdl_P64;
      Type_Desc : Ghdl_Desc_Ptr;
      Resolv_Func : System.Address;
      Resolv_Inst : System.Address)
     return Ghdl_Signal_Ptr;
   procedure Ghdl_Signal_Simple_Assign_P64 (Sign : Ghdl_Signal_Ptr;
                                            Val : Ghdl_P64);
   procedure Ghdl_Signal_Start_Assign_P64 (Sign : Ghdl_Signal_Ptr;
                                           Rej : Std_Time;
                                           Val : Ghdl_P64;
                                           After : Std_Time);
   procedure Ghdl_Signal_Next_Assign_P64 (Sign : Ghdl_Signal_Ptr;
                                          Val : Ghdl_P64;
                                          After : Std_Time);

   function Ghdl_Create_Signal_F64
     (Init_Val : Ghdl_F64;
      Type_Desc : Ghdl_Desc_Ptr;
      Resolv_Func : System.Address;
      Resolv_Inst : System.Address)
     return Ghdl_Signal_Ptr;
   procedure Ghdl_Signal_Simple_Assign_F64 (Sign : Ghdl_Signal_Ptr;
                                            Val : Ghdl_F64);
   procedure Ghdl_Signal_Start_Assign_F64 (Sign : Ghdl_Signal_Ptr;
                                           Rej : Std_Time;
                                           Val : Ghdl_F64;
                                           After : Std_Time);
   procedure Ghdl_Signal_Next_Assign_F64 (Sign : Ghdl_Signal_Ptr;
                                          Val : Ghdl_F64;
                                          After : Std_Time);

   --  Add a driver to SIGN for the current process.
   procedure Ghdl_Process_Add_Driver (Sign : Ghdl_Signal_Ptr);

   --  Used for connexions:
   --  SRC is a source for TARG.
   procedure Ghdl_Signal_Add_Source (Targ : Ghdl_Signal_Ptr;
                                     Src : Ghdl_Signal_Ptr);

   --  Mark the next signals as resolved.
   --  Number of signals is to be computed with TYPE_DESC.
   procedure Ghdl_Signal_Create_Resolution (Type_Desc : Ghdl_Desc_Ptr;
                                            Resolv_Func : System.Address;
                                            Resolv_Inst : System.Address);

   --  The effective value of TARG is the effective value of SRC.
   procedure Ghdl_Signal_Effective_Value (Targ : Ghdl_Signal_Ptr;
                                          Src : Ghdl_Signal_Ptr);

   --  Create a new 'stable (VAL) signal.
   function Ghdl_Create_Stable_Signal (Val : Std_Time) return Ghdl_Signal_Ptr;
   --  Add SIG in the set of prefix for the last created signal.
   procedure Ghdl_Signal_Register_Stable_Prefix (Sig : Ghdl_Signal_Ptr);

   --  Create a new implicitly defined GUARD signal.
   function Ghdl_Signal_Create_Guard (This : System.Address;
                                      Proc : Guard_Func_Acc;
                                      Instance_Name : Ghdl_Instance_Name_Acc)
     return Ghdl_Signal_Ptr;
   --  Add SIG to the list of referenced signals that appear in the guard
   --  expression.
   procedure Ghdl_Signal_Guard_Dependence (Sig : Ghdl_Signal_Ptr);
private
   pragma Export (C, Ghdl_Signal_Name,
                  "__ghdl_signal_name");

   pragma Export (C, Ghdl_Signal_Simple_Assign_Error,
                  "__ghdl_signal_simple_assign_error");
   pragma Export (C, Ghdl_Signal_Start_Assign_Error,
                  "__ghdl_signal_start_assign_error");

   pragma Export (C, Ghdl_Signal_Start_Assign_Null,
                  "__ghdl_signal_start_assign_null");

   pragma Export (C, Ghdl_Create_Signal_B2,
                  "__ghdl_create_signal_b2");
   pragma Export (C, Ghdl_Signal_Simple_Assign_B2,
                  "__ghdl_signal_simple_assign_b2");
   pragma Export (C, Ghdl_Signal_Start_Assign_B2,
                  "__ghdl_signal_start_assign_b2");
   pragma Export (C, Ghdl_Signal_Next_Assign_B2,
                  "__ghdl_signal_next_assign_b2");

   pragma Export (C, Ghdl_Create_Signal_E8,
                  "__ghdl_create_signal_e8");
   pragma Export (C, Ghdl_Signal_Simple_Assign_E8,
                  "__ghdl_signal_simple_assign_e8");
   pragma Export (C, Ghdl_Signal_Start_Assign_E8,
                  "__ghdl_signal_start_assign_e8");
   pragma Export (C, Ghdl_Signal_Next_Assign_E8,
                  "__ghdl_signal_next_assign_e8");

   pragma Export (C, Ghdl_Create_Signal_I32,
                  "__ghdl_create_signal_i32");
   pragma Export (C, Ghdl_Signal_Simple_Assign_I32,
                  "__ghdl_signal_simple_assign_i32");
   pragma Export (C, Ghdl_Signal_Start_Assign_I32,
                  "__ghdl_signal_start_assign_i32");
   pragma Export (C, Ghdl_Signal_Next_Assign_I32,
                  "__ghdl_signal_next_assign_i32");

   pragma Export (C, Ghdl_Create_Signal_P64,
                  "__ghdl_create_signal_p64");
   pragma Export (C, Ghdl_Signal_Simple_Assign_P64,
                  "__ghdl_signal_simple_assign_p64");
   pragma Export (C, Ghdl_Signal_Start_Assign_P64,
                  "__ghdl_signal_start_assign_p64");
   pragma Export (C, Ghdl_Signal_Next_Assign_P64,
                  "__ghdl_signal_next_assign_p64");

   pragma Export (C, Ghdl_Create_Signal_F64,
                  "__ghdl_create_signal_f64");
   pragma Export (C, Ghdl_Signal_Simple_Assign_F64,
                  "__ghdl_signal_simple_assign_f64");
   pragma Export (C, Ghdl_Signal_Start_Assign_F64,
                  "__ghdl_signal_start_assign_f64");
   pragma Export (C, Ghdl_Signal_Next_Assign_F64,
                  "__ghdl_signal_next_assign_f64");

   pragma Export (C, Ghdl_Process_Add_Driver,
                  "__ghdl_process_add_driver");
   pragma Export (C, Ghdl_Signal_Add_Source,
                  "__ghdl_signal_add_source");
   pragma Export (C, Ghdl_Signal_Effective_Value,
                  "__ghdl_signal_effective_value");

   pragma Export (C, Ghdl_Signal_Create_Resolution,
                  "__ghdl_signal_create_resolution");

   pragma Export (C, Ghdl_Create_Stable_Signal,
                  "__ghdl_create_stable_signal");
   pragma Export (C, Ghdl_Signal_Register_Stable_Prefix,
                  "__ghdl_signal_register_stable_prefix");

   pragma Export (C, Ghdl_Signal_Create_Guard,
                  "__ghdl_signal_create_guard");
   pragma Export (C, Ghdl_Signal_Guard_Dependence,
                  "__ghdl_signal_guard_dependence");
end Grt.Signals;
