with Grt.Errors; use Grt.Errors;
with Ada.Unchecked_Conversion;
with System.Storage_Elements; use System.Storage_Elements;
with System.Address_To_Access_Conversions;

package body Grt.Typedesc_Handling is
   Dummy_B2 : Ghdl_B2 := False;
   Dummy_E8 : Ghdl_E8 := 0;
   Dummy_E32 : Ghdl_E32 := 0;
   Dummy_I32 : Ghdl_I32 := 0;
   Dummy_P64 : Ghdl_P64 := 0;
   Dummy_F64 : Ghdl_F64 := 0.0;
   B2_Size : Storage_Offset := Dummy_B2'Size / System.Storage_Unit;
   E8_Size : Storage_Offset := Dummy_E8'Size / System.Storage_Unit;
   E32_Size : Storage_Offset := Dummy_E32'Size / System.Storage_Unit;
   I32_Size : Storage_Offset := Dummy_I32'Size / System.Storage_Unit;
   P64_Size : Storage_Offset := Dummy_P64'Size / System.Storage_Unit;
   F64_Size : Storage_Offset := Dummy_F64'Size / System.Storage_Unit;

   --  Get the length of range ARANGE whose type is described by DESC.
   function Get_Range_Length (Arange : Ghdl_Range_Ptr; Desc : Ghdl_Desc_Ptr)
     return Ghdl_Index_Type
   is
   begin
      case Desc.B2.Kind is
         when Ghdl_Type_B2 =>
            --| Ghdl_Subtype_B2 =>
            return Arange.B2.Len;
         when Ghdl_Type_E8
           | Ghdl_Subtype_E8 =>
            return Arange.E8.Len;
         when Ghdl_Type_E32 =>
            --| Ghdl_Subtype_E32 =>
            return Arange.E32.Len;
         when Ghdl_Type_I32
           | Ghdl_Subtype_I32 =>
            return Arange.I32.Len;
         when others =>
            Internal_Error ("get_range_length");
      end case;
   end Get_Range_Length;

   function Conv is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Ghdl_Range_Ptr);

   --  Get the number of elements for subarray DESC.
   function Get_Subarray_Length (Desc : Ghdl_Desc_Ptr) return Ghdl_Index_Type
   is
      Base : Ghdl_Desc_Ptr;
   begin
      Base := Desc.Sub_Arr.Basetype;
      if Base.Arr.Nbr_Dim /= 1 then
         --  FIXME: todo.
         Internal_Error ("get_subarray_length: subarray");
      else
         return Get_Range_Length (Conv (Desc.Sub_Arr.Trange),
                                  Base.Arr.Indexes (0));
      end if;
   end Get_Subarray_Length;

   function Get_Size (Desc : Ghdl_Desc_Ptr) return Storage_Offset is
   begin
      case Desc.B2.Kind is
         when Ghdl_Type_B2 =>
            return B2_Size;
         when Ghdl_Type_E8 =>
            return E8_Size;
         when Ghdl_Type_I32 =>
            return I32_Size;
         when Ghdl_Subtype_Array =>
            return Get_Size (Desc.Sub_Arr.Basetype.Arr.El_Desc) *
              Storage_Offset (Get_Subarray_Length (Desc));
         when others =>
            Internal_Error ("get_size");
      end case;
   end Get_Size;

   function Get_Nbr_Signals (Desc : Ghdl_Desc_Ptr) return Ghdl_Index_Type is
   begin
      case Desc.B2.Kind is
         when Ghdl_Type_B2
           | Ghdl_Type_E8
           | Ghdl_Type_I32
           | Ghdl_Type_F64
           | Ghdl_Type_P64 =>
            return 1;
         when Ghdl_Subtype_Array =>
            return Get_Subarray_Length (Desc);
         when others =>
            Internal_Error ("get_nbr_signals");
      end case;
   end Get_Nbr_Signals;

   function Allocate (Size : Storage_Offset) return System.Address;
   pragma Import (C, Allocate, "malloc");

   procedure Free (Addr : System.Address);
   pragma Import (C, Free, "free");

   procedure Create_Writer (Res : out Any_Constr_Type; Desc : Ghdl_Desc_Ptr)
   is
      Len : Storage_Offset;
      Ptr : System.Address;
   begin
      Len := Get_Size (Desc);
      Ptr := Allocate (Len);
      Res := Any_Constr_Type'(Desc => Desc, Base => Ptr, Ptr => Ptr);
   end Create_Writer;

   procedure Create_Reader (Res : out Any_Constr_Type; Desc : Ghdl_Desc_Ptr)
     renames Create_Writer;

   procedure Create_Writer (Res : out Any_Constr_Type;
                            Desc : Ghdl_Desc_Ptr; Length : Natural)
   is
      Len : Storage_Offset;
      Ptr : System.Address;
   begin
      Len := Get_Size (Desc);
      Ptr := Allocate (Len * Storage_Offset (Length));
      Res := Any_Constr_Type'(Desc => Desc, Base => Ptr, Ptr => Ptr);
   end Create_Writer;

   procedure Release (Constr : in out Any_Constr_Type) is
   begin
      Free (Constr.Base);
      Constr.Base := System.Null_Address;
      Constr.Ptr := System.Null_Address;
   end Release;

   function Get_Value_Addr (Constr : Any_Constr_Type) return System.Address is
   begin
      return Constr.Base;
   end Get_Value_Addr;

   procedure Align (Addr : in out System.Address; Algn : Storage_Offset);
   pragma Inline (Align);

   procedure Align (Addr : in out System.Address; Algn : Storage_Offset) is
   begin
      Addr := Addr + Algn - 1;
      Addr := Addr - (Addr mod Algn);
   end Align;

   package B2_Conv is new System.Address_To_Access_Conversions
     (Object => Ghdl_B2);

   procedure Write_B2 (Constr : in out Any_Constr_Type; Val : Ghdl_B2) is
   begin
      B2_Conv.To_Pointer (Constr.Ptr).all := Val;
      Constr.Ptr := Constr.Ptr + B2_Size;
   end Write_B2;

   procedure Read_B2 (Constr : in out Any_Constr_Type; Val : out Ghdl_B2) is
   begin
      Val := B2_Conv.To_Pointer (Constr.Ptr).all;
      Constr.Ptr := Constr.Ptr + B2_Size;
   end Read_B2;

   package E8_Conv is new System.Address_To_Access_Conversions
     (Object => Ghdl_E8);

   procedure Write_E8 (Constr : in out Any_Constr_Type; Val : Ghdl_E8) is
   begin
      E8_Conv.To_Pointer (Constr.Ptr).all := Val;
      Constr.Ptr := Constr.Ptr + E8_Size;
   end Write_E8;

   procedure Read_E8 (Constr : in out Any_Constr_Type; Val : out Ghdl_E8) is
   begin
      Val := E8_Conv.To_Pointer (Constr.Ptr).all;
      Constr.Ptr := Constr.Ptr + E8_Size;
   end Read_E8;

   package E32_Conv is new System.Address_To_Access_Conversions
     (Object => Ghdl_E32);

   procedure Write_E32 (Constr : in out Any_Constr_Type; Val : Ghdl_E32) is
   begin
      Align (Constr.Ptr, E32_Size);

      E32_Conv.To_Pointer (Constr.Ptr).all := Val;
      Constr.Ptr := Constr.Ptr + E32_Size;
   end Write_E32;

   procedure Read_E32 (Constr : in out Any_Constr_Type; Val : out Ghdl_E32) is
   begin
      Align (Constr.Ptr, E32_Size);
      Val := E32_Conv.To_Pointer (Constr.Ptr).all;
      Constr.Ptr := Constr.Ptr + E32_Size;
   end Read_E32;

   package I32_Conv is new System.Address_To_Access_Conversions
     (Object => Ghdl_I32);

   procedure Write_I32 (Constr : in out Any_Constr_Type; Val : Ghdl_I32) is
   begin
      Align (Constr.Ptr, I32_Size);

      I32_Conv.To_Pointer (Constr.Ptr).all := Val;
      Constr.Ptr := Constr.Ptr + I32_Size;
   end Write_I32;

   procedure Read_I32 (Constr : in out Any_Constr_Type; Val : out Ghdl_I32) is
   begin
      Align (Constr.Ptr, I32_Size);
      Val := I32_Conv.To_Pointer (Constr.Ptr).all;
      Constr.Ptr := Constr.Ptr + I32_Size;
   end Read_I32;

   package P64_Conv is new System.Address_To_Access_Conversions
     (Object => Ghdl_P64);

   procedure Write_P64 (Constr : in out Any_Constr_Type; Val : Ghdl_P64) is
   begin
      Align (Constr.Ptr, P64_Size);

      P64_Conv.To_Pointer (Constr.Ptr).all := Val;
      Constr.Ptr := Constr.Ptr + P64_Size;
   end Write_P64;

   procedure Read_P64 (Constr : in out Any_Constr_Type; Val : out Ghdl_P64) is
   begin
      Align (Constr.Ptr, P64_Size);
      Val := P64_Conv.To_Pointer (Constr.Ptr).all;
      Constr.Ptr := Constr.Ptr + P64_Size;
   end Read_P64;

   package F64_Conv is new System.Address_To_Access_Conversions
     (Object => Ghdl_F64);

   procedure Write_F64 (Constr : in out Any_Constr_Type; Val : Ghdl_F64) is
   begin
      Align (Constr.Ptr, F64_Size);

      F64_Conv.To_Pointer (Constr.Ptr).all := Val;
      Constr.Ptr := Constr.Ptr + F64_Size;
   end Write_F64;

   procedure Read_F64 (Constr : in out Any_Constr_Type; Val : out Ghdl_F64) is
   begin
      Align (Constr.Ptr, F64_Size);
      Val := F64_Conv.To_Pointer (Constr.Ptr).all;
      Constr.Ptr := Constr.Ptr + F64_Size;
   end Read_F64;

   procedure Write_Value (Constr : in out Any_Constr_Type;
                          Mode : Mode_Type;
                          Val : Value_Union)
   is
   begin
      case Mode is
         when Mode_B2 =>
            Write_B2 (Constr, Val.B2);
         when Mode_E8 =>
            Write_E8 (Constr, Val.E8);
         when Mode_E32 =>
            Write_E32 (Constr, Val.E32);
         when Mode_I32 =>
            Write_I32 (Constr, Val.I32);
         when Mode_P64 =>
            Write_P64 (Constr, Val.P64);
         when Mode_F64 =>
            Write_F64 (Constr, Val.F64);
      end case;
   end Write_Value;

   procedure Read_Value (Constr : in out Any_Constr_Type;
                         Mode : Mode_Type;
                         Res : out Value_Union) is
   begin
      case Mode is
         when Mode_B2 =>
            Read_B2 (Constr, Res.B2);
         when Mode_E8 =>
            Read_E8 (Constr, Res.E8);
         when Mode_E32 =>
            Read_E32 (Constr, Res.E32);
         when Mode_I32 =>
            Read_I32 (Constr, Res.I32);
         when Mode_P64 =>
            Read_P64 (Constr, Res.P64);
         when Mode_F64 =>
            Read_F64 (Constr, Res.F64);
      end case;
   end Read_Value;
end Grt.Typedesc_Handling;
