with Interfaces;
with Interfaces.C_Streams; use Interfaces.C_Streams;
with Ada.Characters.Latin_1; use Ada.Characters.Latin_1;
with System; use System;
with Grt.Errors; use Grt.Errors;
with Grt.Hierarchy; use Grt.Hierarchy;
with Grt.Types; use Grt.Types;
with Grt.Signals; use Grt.Signals;
with Grt.Typedesc; use Grt.Typedesc;
with GNAT.Table;
with Grt.Processes;
with Grt.Disp; use Grt.Disp;

package body Grt.Vcd is
   type String_Acc is access String;

   --  VCD filename.
   Vcd_Filename : String_Acc := null;
   --  Stream corresponding to the VCD filename.
   Vcd_Stream : FILEs;

   --  Index type of the table of vcd variables to dump.
   type Vcd_Index_Type is new Natural;

   --  Time.
   --  This is current_time shifted, since current_time is the time of the
   --  next cycle.
   Vcd_Time : Std_Time;

   --  Return TRUE if OPT is an option for VCD.
   function Vcd_Option (Opt : String) return Boolean
   is
      F : Natural := Opt'First;
   begin
      if Opt'Length < 5 or else Opt (F .. F + 4) /= "--vcd" then
         return False;
      end if;
      if Opt'Length > 6 and then Opt (F + 5) = '=' then
         Vcd_Filename := new String'(Opt (F + 6 .. Opt'Last) & NUL);
         return True;
      else
         return False;
      end if;
   end Vcd_Option;

   procedure Vcd_Help is
   begin
      Put_Line (" --vcd=FILENAME     dump signal values into a VCD file");
   end Vcd_Help;

   procedure Vcd_Put (Str : String)
   is
      R : size_t;
   begin
      R := fwrite (Str'Address, Str'Length, 1, Vcd_Stream);
   end Vcd_Put;

   procedure Vcd_Putc (C : Character)
   is
      R : int;
   begin
      R := fputc (Character'Pos (C), Vcd_Stream);
   end Vcd_Putc;

   procedure Vcd_Newline is
   begin
      Vcd_Putc (LF);
   end Vcd_Newline;

   procedure Vcd_Putline (Str : String) is
   begin
      Vcd_Put (Str);
      Vcd_Newline;
   end Vcd_Putline;

   procedure Vcd_Put (Str : Ghdl_Str_Len_Type)
   is
   begin
      Put_Str_Len (Vcd_Stream, Str);
   end Vcd_Put;

   procedure Vcd_Put_I32 (V : Ghdl_I32)
   is
   begin
      Put_I32 (Vcd_Stream, V);
   end Vcd_Put_I32;

   procedure Vcd_Put_Idcode (N : Vcd_Index_Type)
   is
      Str : String (1 .. 8);
      V, R : Vcd_Index_Type;
      L : Natural;
   begin
      L := 0;
      V := N;
      loop
         R := V mod 93;
         V := V / 93;
         L := L + 1;
         Str (L) := Character'Val (33 + R);
         exit when V = 0;
      end loop;
      Vcd_Put (Str (1 .. L));
   end Vcd_Put_Idcode;

   procedure Vcd_Put_End is
   begin
      Vcd_Putline ("$end");
   end Vcd_Put_End;

   --  Called before elaboration.
   procedure Vcd_Init
   is
      Mode : constant String := "wt" & NUL;
   begin
      if Vcd_Filename = null then
         Vcd_Stream := NULL_Stream;
         return;
      end if;
      if Vcd_Filename.all = "-" & NUL then
         Vcd_Stream := stdout;
      else
         Vcd_Stream := fopen (Vcd_Filename.all'Address, Mode'Address);
         if Vcd_Stream = NULL_Stream then
            Error ("cannot open " & Vcd_Filename (Vcd_Filename'First
                                                  .. Vcd_Filename'Last - 1));
            return;
         end if;
      end if;
      Vcd_Putline ("$date");
      Vcd_Put ("  ");
      declare
         type time_t is new Interfaces.Integer_64;
         Cur_Time : time_t;

         function time (Addr : Address) return time_t;
         pragma Import (C, time);

         function ctime (Timep: Address) return chars;
         pragma Import (C, ctime);

         R : int;
      begin
         Cur_Time := time (Null_Address);
         R := fputs (ctime (Cur_Time'Address), Vcd_Stream);
         -- Note: ctime already append a LF.
      end;
      Vcd_Put_End;
      Vcd_Putline ("$version");
      Vcd_Putline ("  GHDL v0");
      Vcd_Put_End;
      Vcd_Putline ("$timescale");
      Vcd_Putline ("  1 fs");
      Vcd_Put_End;
   end Vcd_Init;

   type Vcd_Var_Kind is (Vcd_Bit, Vcd_Bitvector,
                         Vcd_Stdlogic, Vcd_Stdlogic_Vector,
                         Vcd_Bad);

   type Vcd_Var_Type is record
      Sign : Sig_Table_Index;
      Kind : Vcd_Var_Kind;
      Len : Ghdl_Index_Type;
   end record;

   package Vcd_Table is new GNAT.Table
     (Table_Component_Type => Vcd_Var_Type,
      Table_Index_Type => Vcd_Index_Type,
      Table_Low_Bound => 0,
      Table_Initial => 32,
      Table_Increment => 100);

   procedure Add_Signal (Inst : Ghdl_Instance_Name_Acc)
   is
      Desc : Ghdl_Desc_Ptr;
      Addr : Address;
      Basedesc : Ghdl_Desc_Ptr;
      N : Vcd_Index_Type;
      Vcd_El : Vcd_Var_Type;
   begin
      Desc := Inst.Sig_Type_Desc;
      Addr := Desc.all'Address;
      Vcd_El.Kind := Vcd_Bad;
      case Desc.Kind is
         when Ghdl_Type_B2 =>
            if Addr = Std_Standard_Bit_BTD'Address then
               Vcd_El.Kind := Vcd_Bit;
            end if;
         when Ghdl_Type_E8 =>
            if Addr = Ieee_Std_Logic_1164_Std_Ulogic_BTD'Address then
               Vcd_El.Kind := Vcd_Stdlogic;
            end if;
         when Ghdl_Subtype_E8 =>
            if Addr = Ieee_Std_Logic_1164_Std_Logic_STD'Address then
               Vcd_El.Kind := Vcd_Stdlogic;
            end if;
         when Ghdl_Subtype_Array =>
            Basedesc := Desc.Sub_Arr.Basetype;
            Addr := Basedesc.all'Address;
            if Addr = Std_Standard_Bit_Vector_BTD'Address then
               Vcd_El.Kind := Vcd_Bitvector;
               Vcd_El.Len := To_Ghdl_Range_Ptr (Desc.Sub_Arr.Trange).I32.Len;
            elsif Addr = Ieee_Std_Logic_1164_Std_Logic_Vector_BTD'Address then
               Vcd_El.Kind := Vcd_Stdlogic_Vector;
               Vcd_El.Len := To_Ghdl_Range_Ptr (Desc.Sub_Arr.Trange).I32.Len;
            end if;
         when others =>
            null;
      end case;

      if Vcd_El.Kind = Vcd_Bad then
         Vcd_Put ("$comment ");
         Vcd_Put (Inst.Name.all);
         Vcd_Put (" is ");
         Vcd_Put (Ghdl_Type_Kind'Image (Desc.Kind));
         Vcd_Putc (' ');
         Vcd_Put_End;
         return;
      else
         Vcd_Table.Increment_Last;
         N := Vcd_Table.Last;
         Vcd_El.Sign := Inst.Sig_Index;
         Vcd_Table.Table (N) := Vcd_El;
         Vcd_Put ("$var reg");
         case Vcd_El.Kind is
            when Vcd_Bit
              | Vcd_Stdlogic =>
               Vcd_Put (" 1");
            when Vcd_Bitvector
              | Vcd_Stdlogic_Vector =>
               Vcd_Put (Ghdl_Index_Type'Image (Vcd_El.Len));
            when Vcd_Bad =>
               Internal_Error ("add_signal");
         end case;
         Vcd_Putc (' ');
         Vcd_Put_Idcode (N);
         Vcd_Putc (' ');
         Vcd_Put (Inst.Name.all);
         case Vcd_El.Kind is
            when Vcd_Bit
              | Vcd_Stdlogic =>
               null;
            when Vcd_Bitvector
              | Vcd_Stdlogic_Vector =>
               declare
                  Irange : Ghdl_Range_Ptr;
               begin
                  Irange := To_Ghdl_Range_Ptr (Desc.Sub_Arr.Trange);
                  Vcd_Putc ('[');
                  Vcd_Put_I32 (Irange.I32.Left);
                  Vcd_Putc (':');
                  Vcd_Put_I32 (Irange.I32.Right);
                  Vcd_Putc (']');
               end;
            when Vcd_Bad =>
               null;
         end case;
         Vcd_Putc (' ');
         Vcd_Put_End;
      end if;
   end Add_Signal;

   procedure Vcd_Put_Hierarchy (Instance : Ghdl_Instance_Name_Acc)
   is
      Inst : Ghdl_Instance_Name_Acc;
   begin
      Inst := Instance;
      while Inst /= null loop
         case Inst.Kind is
            when Ghdl_Name_Entity
              | Ghdl_Name_Architecture
              | Ghdl_Name_Block
              | Ghdl_Name_Generate_Iterative
              | Ghdl_Name_Generate_Conditional
              | Ghdl_Name_Component =>
               Vcd_Put ("$scope module ");
               Vcd_Put (Inst.Name.all);
               if Inst.Kind = Ghdl_Name_Generate_Iterative then
                  Vcd_Putc ('(');
                  Disp_Value (Vcd_Stream,
                              Inst.Iter_Addr.all, Inst.Iter_Typedesc);
                  Vcd_Putc (')');
               end if;
               Vcd_Putc (' ');
               Vcd_Put_End;
               Vcd_Put_Hierarchy (Inst.Child);
               Vcd_Put ("$upscope ");
               Vcd_Put_End;
            when Ghdl_Name_Signal =>
               Add_Signal (Inst);
            when others =>
               null;
         end case;
         Inst := Inst.Brother;
      end loop;
   end Vcd_Put_Hierarchy;

   procedure Vcd_Put_Bit (V : Ghdl_B2)
   is
      C : Character;
   begin
      if V then
         C := '1';
      else
         C := '0';
      end if;
      Vcd_Putc (C);
   end Vcd_Put_Bit;

   procedure Vcd_Put_Stdlogic (V : Ghdl_E8)
   is
      type Map_Type is array (Ghdl_E8 range 0 .. 8) of Character;
      --                             "UX01ZWLH-"
      Map_Vlg : constant Map_Type := "xx01zz01x";
      Map_Std : constant Map_Type := "UX01ZWLH-";
   begin
      if V not in Map_Type'Range then
         Vcd_Putc ('?');
      else
         Vcd_Putc (Map_Std (V));
      end if;
   end Vcd_Put_Stdlogic;

   procedure Vcd_Put_Var (I : Vcd_Index_Type)
   is
      Index : Sig_Table_Index;
      V: Vcd_Var_Type renames Vcd_Table.Table (I);
   begin
      Index := V.Sign;
      case V.Kind is
         when Vcd_Bit =>
            Vcd_Put_Bit (Sig_Table.Table (Index).Value.B2);
         when Vcd_Stdlogic =>
            Vcd_Put_Stdlogic (Sig_Table.Table (Index).Value.E8);
         when Vcd_Bitvector =>
            Vcd_Putc ('b');
            for J in 1 .. V.Len loop
               Vcd_Put_Bit (Sig_Table.Table (Index).Value.B2);
               Index := Index + 1;
            end loop;
            Vcd_Putc (' ');
         when Vcd_Stdlogic_Vector =>
            Vcd_Putc ('b');
            for J in 1 .. V.Len loop
               Vcd_Put_Stdlogic (Sig_Table.Table (Index).Value.E8);
               Index := Index + 1;
            end loop;
            Vcd_Putc (' ');
         when Vcd_Bad =>
            null;
      end case;
      Vcd_Put_Idcode (I);
      Vcd_Newline;
   end Vcd_Put_Var;

   --  Check is the value has changed.
   function Vcd_Var_Changed (I : Vcd_Index_Type) return Boolean
   is
      Sig : Sig_Table_Index;
   begin
      Sig := Vcd_Table.Table (I).Sign;
      case Vcd_Table.Table (I).Kind is
         when Vcd_Bit
           | Vcd_Stdlogic =>
            return Sig_Table.Table (Sig).Last_Event = Vcd_Time;
         when Vcd_Bitvector
           | Vcd_Stdlogic_Vector =>
            for J in 1 .. Vcd_Table.Table (I).Len loop
               if Sig_Table.Table (Sig).Last_Event = Vcd_Time then
                  return True;
               end if;
               Sig := Sig + 1;
            end loop;
         when Vcd_Bad =>
            null;
      end case;
      return False;
   end Vcd_Var_Changed;

   procedure Vcd_Put_Time
   is
      Time : String := Std_Time'Image (Vcd_Time);
   begin
      Time (1) := '#';
      Vcd_Putline (Time);
   end Vcd_Put_Time;

   --  Called after elaboration.
   procedure Vcd_Start is
   begin
      --  Do nothing if there is no VCD file to generate.
      if Vcd_Stream = NULL_Stream then
         return;
      end if;
      --  Put hierarchy.
      Vcd_Put_Hierarchy (Top.Child);

      --  At time 0
      Vcd_Time := 0;
   end Vcd_Start;

   --  Called before each non delta cycle.
   procedure Vcd_Cycle is
   begin
      --  Do nothing if there is no VCD file to generate.
      if Vcd_Stream = NULL_Stream then
         return;
      end if;

      --  Disp values.
      Vcd_Put_Time;
      if Vcd_Time = 0 then
         --  Force to disp all values.
         for I in Vcd_Table.First .. Vcd_Table.Last loop
            Vcd_Put_Var (I);
         end loop;
      else
         --  Disp only values changed.
         for I in Vcd_Table.First .. Vcd_Table.Last loop
            if Vcd_Var_Changed (I) then
               Vcd_Put_Var (I);
            end if;
         end loop;
      end if;
      --  Current_time is the time of the next cycle.
      Vcd_Time := Grt.Processes.Current_Time;
   end Vcd_Cycle;

   --  Called at the end of the simulation.
   procedure Vcd_End is
   begin
      null;
   end Vcd_End;
end Grt.Vcd;
