with Types; use Types;
with Scan; use Scan;
with Tokens; use Tokens;
with Errorout; use Errorout;
with Name_Table;
with Std_Names; use Std_Names;
with Std_Package;
with Flags;

package body Iirs_Utils is
   -- Transform the current token into an iir literal.
   -- The current token must be either a character or an identifier.
   function Current_Text return Iir is
      Res: Iir;
   begin
      case Current_Token is
         when Tok_Identifier =>
            Res := Create_Iir (Iir_Kind_Simple_Name);
         when Tok_String =>
            Res := Create_Iir (Iir_Kind_String_Literal);
         when Tok_Character =>
            Res := Create_Iir (Iir_Kind_Character_Literal);
         when others =>
            raise Internal_Error;
      end case;
      Set_Identifier (Res, Current_Identifier);
      Invalidate_Current_Identifier;
      Invalidate_Current_Token;
      Set_Location (Res);
      return Res;
   end Current_Text;

   procedure Set_Location (An_Iir: Iir) is
      use Iirs;
   begin
      Set_Location (An_Iir, Get_Token_Location);
   end Set_Location;

   function Get_Operator_Name (Op : Iir) return Name_Id is
   begin
      case Get_Kind (Op) is
         when Iir_Kind_And_Operator =>
            return Name_And;
         when Iir_Kind_Or_Operator =>
            return Name_Or;
         when Iir_Kind_Nand_Operator =>
            return Name_Nand;
         when Iir_Kind_Nor_Operator =>
            return Name_Nor;
         when Iir_Kind_Xor_Operator =>
            return Name_Xor;
         when Iir_Kind_Xnor_Operator =>
            return Name_Xnor;
         when Iir_Kind_Equality_Operator =>
            return Name_Op_Equality;
         when Iir_Kind_Inequality_Operator =>
            return Name_Op_Inequality;
         when Iir_Kind_Less_Than_Operator =>
            return Name_Op_Less;
         when Iir_Kind_Less_Than_Or_Equal_Operator =>
            return Name_Op_Less_Equal;
         when Iir_Kind_Greater_Than_Operator =>
            return Name_Op_Greater;
         when Iir_Kind_Greater_Than_Or_Equal_Operator =>
            return Name_Op_Greater_Equal;
         when Iir_Kind_Sll_Operator =>
            return Name_Sll;
         when Iir_Kind_Sla_Operator =>
            return Name_Sla;
         when Iir_Kind_Srl_Operator =>
            return Name_Srl;
         when Iir_Kind_Sra_Operator =>
            return Name_Sra;
         when Iir_Kind_Rol_Operator =>
            return Name_Rol;
         when Iir_Kind_Ror_Operator =>
            return Name_Ror;
         when Iir_Kind_Addition_Operator =>
            return Name_Op_Plus;
         when Iir_Kind_Substraction_Operator =>
            return Name_Op_Minus;
         when Iir_Kind_Concatenation_Operator =>
            return Name_Op_Concatenation;
         when Iir_Kind_Multiplication_Operator =>
            return Name_Op_Mul;
         when Iir_Kind_Division_Operator =>
            return Name_Op_Div;
         when Iir_Kind_Modulus_Operator =>
            return Name_Mod;
         when Iir_Kind_Remainder_Operator =>
            return Name_Rem;
         when Iir_Kind_Exponentiation_Operator =>
            return Name_Op_Exp;
         when Iir_Kind_Not_Operator =>
            return Name_Not;
         when Iir_Kind_Negation_Operator =>
            return Name_Op_Minus;
         when Iir_Kind_Identity_Operator =>
            return Name_Op_Plus;
         when Iir_Kind_Absolute_Operator =>
            return Name_Abs;
         when others =>
            raise Internal_Error;
      end case;
   end Get_Operator_Name;

   function Get_Longuest_Static_Prefix (Expr: Iir) return Iir is
      Adecl: Iir;
   begin
      Adecl := Expr;
      loop
         case Get_Kind (Adecl) is
            when Iir_Kind_Variable_Declaration
              | Iir_Kind_Variable_Interface_Declaration =>
               return Adecl;
            when Iir_Kind_Constant_Declaration
              | Iir_Kind_Constant_Interface_Declaration =>
               return Adecl;
            when Iir_Kind_Signal_Declaration
              | Iir_Kind_Guard_Signal_Declaration
              | Iir_Kind_Signal_Interface_Declaration =>
               return Adecl;
            when Iir_Kind_Slice_Name
              | Iir_Kind_Indexed_Name
              | Iir_Kind_Selected_Element =>
               if Get_Name_Staticness (Adecl) >= Globally then
                  return Adecl;
               else
                  Adecl := Get_Prefix (Adecl);
               end if;
            when Iir_Kind_Type_Conversion =>
               return Null_Iir;
            when others =>
               Error_Kind ("get_longuest_static_prefix", Adecl);
         end case;
      end loop;
   end Get_Longuest_Static_Prefix;

   function Get_Object_Prefix (Decl: Iir) return Iir is
      Adecl: Iir;
   begin
      Adecl := Decl;
      loop
         case Get_Kind (Adecl) is
            when Iir_Kind_Variable_Declaration
              | Iir_Kind_Variable_Interface_Declaration
              | Iir_Kind_Constant_Declaration
              | Iir_Kind_Constant_Interface_Declaration
              | Iir_Kind_Signal_Declaration
              | Iir_Kind_Guard_Signal_Declaration
              | Iir_Kind_Signal_Interface_Declaration
              | Iir_Kind_File_Declaration
              | Iir_Kind_File_Interface_Declaration
              | Iir_Kind_Iterator_Declaration =>
               return Adecl;
            when Iir_Kind_Alias_Declaration =>
               Adecl := Get_Name (Adecl);
            when Iir_Kind_Indexed_Name
              | Iir_Kind_Slice_Name
              | Iir_Kind_Selected_Element
              | Iir_Kind_Selected_By_All_Name
              | Iir_Kind_Implicit_Dereference =>
               Adecl := Get_Prefix (Adecl);
            when Iir_Kinds_Literal
              | Iir_Kind_Enumeration_Literal
              | Iir_Kinds_Monadic_Operator
              | Iir_Kinds_Dyadic_Operator
              | Iir_Kind_Function_Call
              | Iir_Kind_Qualified_Expression
              | Iir_Kind_Type_Conversion
              | Iir_Kind_Allocator_By_Expression
              | Iir_Kind_Allocator_By_Subtype
              | Iir_Kinds_Attribute =>
               return Adecl;
            when others =>
               Error_Kind ("get_object_prefix", Adecl);
         end case;
      end loop;
   end Get_Object_Prefix;

   function Find_Name_In_List (List: Iir_List; Lit: Name_Id) return Iir is
      El: Iir;
      Ident: Name_Id;
   begin
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = null;
         Ident := Get_Identifier (El);
         if Ident = Lit then
            return El;
         end if;
      end loop;
      return Null_Iir;
   end Find_Name_In_List;

   function Is_In_List (List : Iir_List; El : Iir) return Boolean
   is
      List_El : Iir;
   begin
      for I in Natural loop
         List_El := Get_Nth_Element (List, I);
         exit when List_El = null;
         if List_El = El then
            return True;
         end if;
      end loop;
      return False;
   end Is_In_List;

   procedure Add_Dependence (Target: Iir_Design_Unit; Unit: Iir) is
   begin
      Add_Element (Get_Dependence_List (Target), Unit);
   end Add_Dependence;

   function Get_Subprogram_Body (Subprg: Iir) return Iir_List is
   begin
      case Get_Kind (Subprg) is
         when Iir_Kind_Function_Body
           | Iir_Kind_Function_Spec_Body
           | Iir_Kind_Procedure_Body
           | Iir_Kind_Procedure_Spec_Body =>
            return Subprg;
         when Iir_Kind_Function_Specification
           | Iir_Kind_Procedure_Specification =>
            return Iirs.Get_Subprogram_Body (Subprg);
         when Iir_Kind_Implicit_Function_Declaration
           | Iir_Kind_Implicit_Procedure_Declaration =>
            return null;
         when others =>
            raise Internal_Error;
      end case;
   end Get_Subprogram_Body;

   function Get_Subprogram_Spec (Subprg: Iir) return Iir is
   begin
      case Get_Kind (Subprg) is
         when Iir_Kind_Function_Body
           | Iir_Kind_Procedure_Body =>
            return Iirs.Get_Subprogram_Specification (Subprg);
         when Iir_Kind_Function_Specification
           | Iir_Kind_Function_Spec_Body
           | Iir_Kind_Procedure_Specification
           | Iir_Kind_Procedure_Spec_Body =>
            return Subprg;
         when others =>
            raise Internal_Error;
      end case;
   end Get_Subprogram_Spec;

   procedure Clear_Instantiation_Configuration
     (List : Iir_Concurrent_Statement_List)
   is
      El : Iir;
   begin
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = null;
         case Get_Kind (El) is
            when Iir_Kind_Component_Instantiation_Statement =>
               Set_Component_Configuration (El, Null_Iir);
            when Iir_Kind_Generate_Statement =>
               Set_Generate_Block_Configuration (El, Null_Iir);
            when Iir_Kind_Block_Statement =>
               Set_Block_Block_Configuration (El, Null_Iir);
            when others =>
               null;
         end case;
      end loop;
   end Clear_Instantiation_Configuration;

   --  Get identifier of NODE as a string.
   function Get_String (Node : Iir) return String is
   begin
      return Name_Table.Image (Iirs.Get_Identifier (Node));
   end Get_String;

   --  LRM93 7.2.2
   --  A discrete array is a one-dimensional array whose elements are of a
   --  discrete type.
   function Is_Discrete_Array (Def : Iir) return Boolean
   is
   begin
      case Get_Kind (Def) is
         when Iir_Kind_Array_Type_Definition
           | Iir_Kind_Array_Subtype_Definition =>
            null;
         when others =>
            raise Internal_Error;
            -- return False;
      end case;
      if Get_Nbr_Elements (Get_Index_List (Def)) /= 1 then
         return False;
      end if;
      if Get_Kind (Get_Element_Subtype (Def))
        not in Iir_Kinds_Discrete_Type_Definition
      then
         return False;
      end if;
      return True;
   end Is_Discrete_Array;

   procedure Create_Implicit_File_Primitives
     (Decl : Iir_Type_Declaration; Type_Definition : Iir_File_Type_Definition)
   is
      Type_Mark: Iir;
      Proc: Iir_Implicit_Procedure_Declaration;
      Func: Iir_Implicit_Function_Declaration;
      Interface: Iir;
      Loc : Location_Type;
      File_Interface_Kind : Iir_Kind;
   begin
      Type_Mark := Get_Type_Mark (Type_Definition);
      Loc := Get_Location (Decl);

      if Flags.Vhdl_Std >= Vhdl_93c then
         for I in 1 .. 2 loop
            --  Create the implicit file_open (form 1) declaration.
            --  Create the implicit file_open (form 2) declaration.
            Proc := Create_Iir (Iir_Kind_Implicit_Procedure_Declaration);
            Set_Location (Proc, Loc);
            Set_Identifier (Proc, Std_Names.Name_File_Open);
            case I is
               when 1 =>
                  Set_Implicit_Definition (Proc, Iir_Predefined_File_Open);
               when 2 =>
                  Set_Implicit_Definition (Proc,
                                           Iir_Predefined_File_Open_Status);
                  --  status : out file_open_status.
                  Interface :=
                    Create_Iir (Iir_Kind_Variable_Interface_Declaration);
                  Set_Location (Interface, Loc);
                  Set_Identifier (Interface, Std_Names.Name_Status);
                  Set_Type (Interface,
                            Std_Package.File_Open_Status_Type_Definition);
                  Set_Mode (Interface, Iir_Out_Mode);
                  Append_Element (Get_Interface_Declaration (Proc), Interface);
            end case;
            --  File F : FT
            Interface := Create_Iir (Iir_Kind_File_Interface_Declaration);
            Set_Location (Interface, Loc);
            Set_Identifier (Interface, Std_Names.Name_F);
            Set_Type (Interface, Type_Definition);
            Set_Mode (Interface, Iir_Inout_Mode);
            Append_Element (Get_Interface_Declaration (Proc), Interface);
            --  External_Name : in STRING
            Interface := Create_Iir (Iir_Kind_Constant_Interface_Declaration);
            Set_Location (Interface, Loc);
            Set_Identifier (Interface, Std_Names.Name_External_Name);
            Set_Type (Interface, Std_Package.String_Type_Definition);
            Set_Mode (Interface, Iir_In_Mode);
            Append_Element (Get_Interface_Declaration (Proc), Interface);
            --  Open_Kind : in File_Open_Kind := Read_Mode.
            Interface := Create_Iir (Iir_Kind_Constant_Interface_Declaration);
            Set_Location (Interface, Loc);
            Set_Identifier (Interface, Std_Names.Name_Open_Kind);
            Set_Type (Interface, Std_Package.File_Open_Kind_Type_Definition);
            Set_Mode (Interface, Iir_In_Mode);
            Set_Default_Value (Interface,
                               Std_Package.File_Open_Kind_Read_Mode);
            Append_Element (Get_Interface_Declaration (Proc), Interface);
            -- Add it to the list.
            Append_Element (Get_Implicit_Functions_List (Decl), Proc);
         end loop;

         --  Create the implicit file_close declaration.
         Proc := Create_Iir (Iir_Kind_Implicit_Procedure_Declaration);
         Set_Identifier (Proc, Std_Names.Name_File_Close);
         Set_Implicit_Definition (Proc, Iir_Predefined_File_Close);
         Interface := Create_Iir (Iir_Kind_File_Interface_Declaration);
         Set_Identifier (Interface, Std_Names.Name_F);
         Set_Type (Interface, Type_Definition);
         Set_Mode (Interface, Iir_Inout_Mode);
         Append_Element (Get_Interface_Declaration (Proc), Interface);
         -- Add it to the list.
         Append_Element (Get_Implicit_Functions_List (Decl), Proc);
      end if;

      if Flags.Vhdl_Std = Vhdl_87 then
         File_Interface_Kind := Iir_Kind_Variable_Interface_Declaration;
      else
         File_Interface_Kind := Iir_Kind_File_Interface_Declaration;
      end if;

      -- Create the implicit procedure read declaration.
      Proc := Create_Iir (Iir_Kind_Implicit_Procedure_Declaration);
      Set_Identifier (Proc, Std_Names.Name_Read);
      Interface := Create_Iir (File_Interface_Kind);
      Set_Identifier (Interface, Std_Names.Name_F);
      Set_Type (Interface, Type_Definition);
      Set_Mode (Interface, Iir_Inout_Mode);
      Append_Element (Get_Interface_Declaration (Proc), Interface);
      Interface := Create_Iir (Iir_Kind_Variable_Interface_Declaration);
      Set_Identifier (Interface, Std_Names.Name_Value);
      Set_Type (Interface, Type_Mark);
      Set_Mode (Interface, Iir_Out_Mode);
      Append_Element (Get_Interface_Declaration (Proc), Interface);
      case Get_Kind (Type_Mark) is
         when Iir_Kind_Array_Type_Definition
           | Iir_Kind_Array_Subtype_Definition =>
            Interface := Create_Iir (Iir_Kind_Variable_Interface_Declaration);
            Set_Identifier (Interface, Std_Names.Name_Length);
            Set_Type (Interface, Std_Package.Natural_Subtype_Definition);
            Set_Mode (Interface, Iir_Out_Mode);
            Append_Element (Get_Interface_Declaration (Proc), Interface);
            Set_Implicit_Definition (Proc, Iir_Predefined_Read_Length);
         when others =>
            Set_Implicit_Definition (Proc, Iir_Predefined_Read);
      end case;
      -- Add it to the list.
      Append_Element (Get_Implicit_Functions_List (Decl), Proc);

      -- Create the implicit procedure write declaration.
      Proc := Create_Iir (Iir_Kind_Implicit_Procedure_Declaration);
      Set_Identifier (Proc, Std_Names.Name_Write);
      Interface := Create_Iir (File_Interface_Kind);
      Set_Identifier (Interface, Std_Names.Name_F);
      Set_Type (Interface, Type_Definition);
      Set_Mode (Interface, Iir_Inout_Mode);
      Append_Element (Get_Interface_Declaration (Proc), Interface);
      Interface := Create_Iir (Iir_Kind_Constant_Interface_Declaration);
      Set_Identifier (Interface, Std_Names.Name_Value);
      Set_Type (Interface, Type_Mark);
      Set_Mode (Interface, Iir_In_Mode);
      Append_Element (Get_Interface_Declaration (Proc), Interface);
      Set_Implicit_Definition (Proc, Iir_Predefined_Write);
      -- Add it to the list.
      Append_Element (Get_Implicit_Functions_List (Decl), Proc);

      -- Create the implicit function endfile declaration.
      Func := Create_Iir (Iir_Kind_Implicit_Function_Declaration);
      Set_Identifier (Func, Std_Names.Name_Endfile);
      Interface := Create_Iir (File_Interface_Kind);
      Set_Identifier (Interface, Std_Names.Name_F);
      Set_Type (Interface, Type_Definition);
      Set_Mode (Interface, Iir_Inout_Mode);
      Set_Interface_Declaration
        (Func, Create_Iir (Iir_Kind_Interface_List));
      Append_Element (Get_Interface_Declaration (Func), Interface);
      Set_Return_Type (Func, Std_Package.Boolean_Type_Definition);
      Set_Implicit_Definition (Func, Iir_Predefined_Endfile);
      -- Add it to the list.
      Append_Element (Get_Implicit_Functions_List (Decl), Func);
   end Create_Implicit_File_Primitives;

   procedure Create_Implicit_Operations
     (Decl : Iir_Type_Declaration; Is_Std_Standard : Boolean := False)
   is
      Interface : Iir_Constant_Interface_Declaration;
      Binary_List : Iir_Interface_List;
      Unary_List : Iir_Interface_List;
      List : Iir_Interface_List;
      Type_Definition : Iir;
      Element_Interface : Iir;

      procedure Add_Operation
        (Name : Name_Id;
         Def : Iir_Predefined_Functions;
         Interface_List : Iir_Interface_List;
         Return_Type : Iir)
      is
         Operation : Iir_Implicit_Function_Declaration;
      begin
         Operation := Create_Iir (Iir_Kind_Implicit_Function_Declaration);
         Location_Copy (Operation, Decl);
         Set_Interface_Declaration (Operation, Interface_List);
         Set_Return_Type (Operation, Return_Type);
         Set_Implicit_Definition (Operation, Def);
         Set_Identifier (Operation, Name);
         Append_Element (Get_Implicit_Functions_List (Decl), Operation);
      end Add_Operation;

      procedure Add_Relational (Name : Name_Id; Def : Iir_Predefined_Functions)
      is
      begin
         Add_Operation
           (Name, Def, Binary_List, Std_Package.Boolean_Type_Definition);
      end Add_Relational;

      procedure Add_Binary (Name : Name_Id; Def : Iir_Predefined_Functions) is
      begin
         Add_Operation (Name, Def, Binary_List, Type_Definition);
      end Add_Binary;

      procedure Add_Unary (Name : Name_Id; Def : Iir_Predefined_Functions) is
      begin
         Add_Operation (Name, Def, Unary_List, Type_Definition);
      end Add_Unary;
   begin
      Type_Definition := Get_Base_Type (Get_Type (Decl));
      if Get_Kind (Type_Definition) /= Iir_Kind_File_Type_Definition then
         Interface := Create_Iir (Iir_Kind_Constant_Interface_Declaration);
         Location_Copy (Interface, Decl);
         Set_Identifier (Interface, Null_Identifier);
         Set_Mode (Interface, Iir_In_Mode);
         Set_Anonymous_Interface (Decl, Interface);
         Set_Type (Interface, Type_Definition);
         Binary_List := Create_Iir (Iir_Kind_Interface_List);
         Append_Element (Binary_List, Interface);
         Append_Element (Binary_List, Interface);
      end if;

      case Get_Kind (Type_Definition) is
         when Iir_Kind_Enumeration_Type_Definition =>
            Add_Relational (Name_Op_Equality, Iir_Predefined_Enum_Equality);
            Add_Relational
              (Name_Op_Inequality, Iir_Predefined_Enum_Inequality);
            Add_Relational (Name_Op_Greater, Iir_Predefined_Enum_Greater);
            Add_Relational
              (Name_Op_Greater_Equal, Iir_Predefined_Enum_Greater_Equal);
            Add_Relational (Name_Op_Less, Iir_Predefined_Enum_Less);
            Add_Relational
              (Name_Op_Less_Equal, Iir_Predefined_Enum_Less_Equal);

         when Iir_Kind_Array_Type_Definition
           | Iir_Kind_Array_Subtype_Definition =>
            declare
               Element_Type : Iir;
            begin
               Add_Relational
                 (Name_Op_Equality, Iir_Predefined_Array_Equality);
               Add_Relational
                 (Name_Op_Inequality, Iir_Predefined_Array_Inequality);
               if Is_Discrete_Array (Type_Definition) then
                  Add_Relational
                    (Name_Op_Greater, Iir_Predefined_Array_Greater);
                  Add_Relational
                    (Name_Op_Greater_Equal,
                     Iir_Predefined_Array_Greater_Equal);
                  Add_Relational
                    (Name_Op_Less, Iir_Predefined_Array_Less);
                  Add_Relational
                    (Name_Op_Less_Equal, Iir_Predefined_Array_Less_Equal);
               end if;

               Element_Type :=
                 Get_Base_Type (Get_Element_Subtype (Type_Definition));
               Element_Interface := Get_Anonymous_Interface
                 (Get_Type_Declarator (Element_Type));
               List := Create_Iir (Iir_Kind_Interface_List);
               Append_Element (List, Interface);
               Append_Element (List, Interface);
               Add_Operation (Name_Op_Concatenation,
                              Iir_Predefined_Array_Array_Concat,
                              List,
                              Type_Definition);

               List := Create_Iir (Iir_Kind_Interface_List);
               Append_Element (List, Element_Interface);
               Append_Element (List, Interface);
               Add_Operation (Name_Op_Concatenation,
                              Iir_Predefined_Element_Array_Concat,
                              List,
                              Type_Definition);

               List := Create_Iir (Iir_Kind_Interface_List);
               Append_Element (List, Interface);
               Append_Element (List, Element_Interface);
               Add_Operation (Name_Op_Concatenation,
                              Iir_Predefined_Array_Element_Concat,
                              List,
                              Type_Definition);
               List := Create_Iir (Iir_Kind_Interface_List);
               Append_Element (List, Element_Interface);
               Append_Element (List, Element_Interface);
               Add_Operation (Name_Op_Concatenation,
                              Iir_Predefined_Element_Element_Concat,
                              List,
                              Type_Definition);

               if Is_Discrete_Array (Type_Definition) then
                  if Element_Type = Std_Package.Boolean_Type_Definition then
                     Unary_List := Create_Iir (Iir_Kind_Interface_List);
                     Append_Element (Unary_List, Interface);

                     Add_Binary (Name_And, Iir_Predefined_Boolean_Array_And);
                     Add_Binary (Name_Or, Iir_Predefined_Boolean_Array_Or);
                     Add_Binary (Name_Nand, Iir_Predefined_Boolean_Array_Nand);
                     Add_Binary (Name_Nor, Iir_Predefined_Boolean_Array_Nor);
                     Add_Binary (Name_Xor, Iir_Predefined_Boolean_Array_Xor);
                     if Flags.Vhdl_Std > Vhdl_87 then
                        Add_Binary
                          (Name_Xnor, Iir_Predefined_Boolean_Array_Xnor);
                     end if;
                     Add_Unary (Name_Not, Iir_Predefined_Boolean_Array_Not);
                  elsif Element_Type = Std_Package.Bit_Type_Definition then
                     Unary_List := Create_Iir (Iir_Kind_Interface_List);
                     Append_Element (Unary_List, Interface);

                     Add_Binary (Name_And, Iir_Predefined_Bit_Array_And);
                     Add_Binary (Name_Or, Iir_Predefined_Bit_Array_Or);
                     Add_Binary (Name_Nand, Iir_Predefined_Bit_Array_Nand);
                     Add_Binary (Name_Nor, Iir_Predefined_Bit_Array_Nor);
                     Add_Binary (Name_Xor, Iir_Predefined_Bit_Array_Xor);
                     if Flags.Vhdl_Std > Vhdl_87 then
                        Add_Binary (Name_Xnor, Iir_Predefined_Bit_Array_Xnor);
                     end if;
                     Add_Unary (Name_Not, Iir_Predefined_Bit_Array_Not);
                  end if;
               end if;
            end;

         when Iir_Kind_Access_Type_Definition =>
            Add_Relational (Name_Op_Equality, Iir_Predefined_Access_Equality);
            Add_Relational
              (Name_Op_Inequality, Iir_Predefined_Access_Inequality);
            declare
               Deallocate_Proc: Iir_Implicit_Procedure_Declaration;
               Var_Interface: Iir_Variable_Interface_Declaration;
            begin
               Deallocate_Proc :=
                 Create_Iir (Iir_Kind_Implicit_Procedure_Declaration);
               Set_Identifier (Deallocate_Proc, Std_Names.Name_Deallocate);
               Var_Interface :=
                 Create_Iir (Iir_Kind_Variable_Interface_Declaration);
               Set_Identifier (Var_Interface, Std_Names.Name_P);
               Set_Type (Var_Interface, Type_Definition);
               Set_Mode (Var_Interface, Iir_Inout_Mode);
               Set_Passive_State (Deallocate_Proc, True);
               Set_Purity_State (Deallocate_Proc, False);
               Set_Wait_State (Deallocate_Proc, False);
               Set_Implicit_Definition
                 (Deallocate_Proc, Iir_Predefined_Deallocate);

               Append_Element (Get_Interface_Declaration (Deallocate_Proc),
                               Var_Interface);
               Append_Element (Get_Implicit_Functions_List (Decl),
                               Deallocate_Proc);
            end;

         when Iir_Kind_Record_Type_Definition =>
            Add_Relational (Name_Op_Equality, Iir_Predefined_Record_Equality);
            Add_Relational
              (Name_Op_Inequality, Iir_Predefined_Record_Inequality);

         when Iir_Kind_Integer_Type_Definition =>
            Unary_List := Create_Iir (Iir_Kind_Interface_List);
            Append_Element (Unary_List, Interface);
            Add_Relational (Name_Op_Equality, Iir_Predefined_Integer_Equality);
            Add_Relational
              (Name_Op_Inequality, Iir_Predefined_Integer_Inequality);
            Add_Relational (Name_Op_Greater, Iir_Predefined_Integer_Greater);
            Add_Relational
              (Name_Op_Greater_Equal, Iir_Predefined_Integer_Greater_Equal);
            Add_Relational (Name_Op_Less, Iir_Predefined_Integer_Less);
            Add_Relational
              (Name_Op_Less_Equal, Iir_Predefined_Integer_Less_Equal);

            Add_Binary (Name_Op_Plus, Iir_Predefined_Integer_Plus);
            Add_Binary (Name_Op_Minus, Iir_Predefined_Integer_Minus);

            Add_Unary (Name_Op_Minus, Iir_Predefined_Integer_Negation);
            Add_Unary (Name_Op_Plus, Iir_Predefined_Integer_Identity);

            Add_Binary (Name_Op_Mul, Iir_Predefined_Integer_Mul);
            Add_Binary (Name_Op_Div, Iir_Predefined_Integer_Div);
            Add_Binary (Name_Mod, Iir_Predefined_Integer_Mod);
            Add_Binary (Name_Rem, Iir_Predefined_Integer_Rem);

            Add_Unary (Name_Abs, Iir_Predefined_Integer_Absolute);

            if not Is_Std_Standard
              or else Decl /= Std_Package.Universal_Integer_Type
            then
               List := Create_Iir (Iir_Kind_Interface_List);
               Append_Element (List, Interface);
               Append_Element
                 (List, Get_Anonymous_Interface (Std_Package.Integer_Type));
               Add_Operation (Name_Op_Exp, Iir_Predefined_Integer_Exp,
                              List, Type_Definition);
            end if;

         when Iir_Kind_Floating_Type_Definition =>
            Unary_List := Create_Iir (Iir_Kind_Interface_List);
            Append_Element (Unary_List, Interface);
            Add_Relational
              (Name_Op_Equality, Iir_Predefined_Floating_Equality);
            Add_Relational
              (Name_Op_Inequality, Iir_Predefined_Floating_Inequality);
            Add_Relational
              (Name_Op_Greater, Iir_Predefined_Floating_Greater);
            Add_Relational
              (Name_Op_Greater_Equal, Iir_Predefined_Floating_Greater_Equal);
            Add_Relational
              (Name_Op_Less, Iir_Predefined_Floating_Less);
            Add_Relational
              (Name_Op_Less_Equal, Iir_Predefined_Floating_Less_Equal);

            Add_Binary (Name_Op_Plus, Iir_Predefined_Floating_Plus);
            Add_Binary (Name_Op_Minus, Iir_Predefined_Floating_Minus);

            Add_Unary (Name_Op_Minus, Iir_Predefined_Floating_Negation);
            Add_Unary (Name_Op_Plus, Iir_Predefined_Floating_Identity);

            Add_Binary (Name_Op_Mul, Iir_Predefined_Floating_Mul);
            Add_Binary (Name_Op_Div, Iir_Predefined_Floating_Div);

            Add_Unary (Name_Abs, Iir_Predefined_Floating_Absolute);

            if not Is_Std_Standard
              or else Decl /= Std_Package.Universal_Real_Type
            then
               List := Create_Iir (Iir_Kind_Interface_List);
               Append_Element (List, Interface);
               Append_Element
                 (List, Get_Anonymous_Interface (Std_Package.Integer_Type));
               Add_Operation (Name_Op_Exp, Iir_Predefined_Floating_Exp,
                              List, Type_Definition);
            end if;

         when Iir_Kind_Physical_Type_Definition =>
            Unary_List := Create_Iir (Iir_Kind_Interface_List);
            Append_Element (Unary_List, Interface);
            Add_Relational
              (Name_Op_Equality, Iir_Predefined_Physical_Equality);
            Add_Relational
              (Name_Op_Inequality, Iir_Predefined_Physical_Inequality);
            Add_Relational
              (Name_Op_Greater, Iir_Predefined_Physical_Greater);
            Add_Relational
              (Name_Op_Greater_Equal, Iir_Predefined_Physical_Greater_Equal);
            Add_Relational
              (Name_Op_Less, Iir_Predefined_Physical_Less);
            Add_Relational
              (Name_Op_Less_Equal, Iir_Predefined_Physical_Less_Equal);

            Add_Binary (Name_Op_Plus, Iir_Predefined_Physical_Plus);
            Add_Binary (Name_Op_Minus, Iir_Predefined_Physical_Minus);

            Add_Unary (Name_Op_Minus, Iir_Predefined_Physical_Negation);
            Add_Unary (Name_Op_Plus, Iir_Predefined_Physical_Identity);

            List := Create_Iir (Iir_Kind_Interface_List);
            Append_Element (List, Interface);
            Append_Element
              (List, Get_Anonymous_Interface (Std_Package.Integer_Type));
            Add_Operation (Name_Op_Mul, Iir_Predefined_Physical_Integer_Mul,
                           List, Type_Definition);
            Add_Operation (Name_Op_Div, Iir_Predefined_Physical_Integer_Div,
                           List, Type_Definition);

            List := Create_Iir (Iir_Kind_Interface_List);
            Append_Element
              (List, Get_Anonymous_Interface (Std_Package.Integer_Type));
            Append_Element (List, Interface);
            Add_Operation (Name_Op_Mul, Iir_Predefined_Integer_Physical_Mul,
                           List, Type_Definition);

            List := Create_Iir (Iir_Kind_Interface_List);
            Append_Element (List, Interface);
            Append_Element
              (List, Get_Anonymous_Interface (Std_Package.Real_Type));
            Add_Operation (Name_Op_Mul, Iir_Predefined_Physical_Real_Mul,
                           List, Type_Definition);
            Add_Operation (Name_Op_Div, Iir_Predefined_Physical_Real_Div,
                           List, Type_Definition);

            List := Create_Iir (Iir_Kind_Interface_List);
            Append_Element
              (List, Get_Anonymous_Interface (Std_Package.Real_Type));
            Append_Element (List, Interface);
            Add_Operation (Name_Op_Mul, Iir_Predefined_Real_Physical_Mul,
                           List, Type_Definition);

            Add_Operation (Name_Op_Div, Iir_Predefined_Physical_Physical_Div,
                           Binary_List,
                           Std_Package.Universal_Integer_Type_Definition);

            Add_Unary (Name_Abs, Iir_Predefined_Physical_Absolute);

         when Iir_Kind_File_Type_Definition =>
            Create_Implicit_File_Primitives (Decl, Type_Definition);

         when others =>
            Error_Kind ("create_predefined_operations", Type_Definition);
      end case;

      if not Is_Std_Standard then
         return;
      end if;
      if Decl = Std_Package.Boolean_Type then
         Unary_List := Create_Iir (Iir_Kind_Interface_List);
         Append_Element (Unary_List, Interface);
         Add_Binary (Name_And, Iir_Predefined_Boolean_And);
         Add_Binary (Name_Or, Iir_Predefined_Boolean_Or);
         Add_Binary (Name_Nand, Iir_Predefined_Boolean_Nand);
         Add_Binary (Name_Nor, Iir_Predefined_Boolean_Nor);
         Add_Binary (Name_Xor, Iir_Predefined_Boolean_Xor);
         if Flags.Vhdl_Std > Vhdl_87 then
            Add_Binary (Name_Xnor, Iir_Predefined_Boolean_Xnor);
         end if;
         Add_Unary (Name_Not, Iir_Predefined_Boolean_Not);
      elsif Decl = Std_Package.Bit_Type then
         Unary_List := Create_Iir (Iir_Kind_Interface_List);
         Append_Element (Unary_List, Interface);
         Add_Binary (Name_And, Iir_Predefined_Bit_And);
         Add_Binary (Name_Or, Iir_Predefined_Bit_Or);
         Add_Binary (Name_Nand, Iir_Predefined_Bit_Nand);
         Add_Binary (Name_Nor, Iir_Predefined_Bit_Nor);
         Add_Binary (Name_Xor, Iir_Predefined_Bit_Xor);
         if Flags.Vhdl_Std > Vhdl_87 then
            Add_Binary (Name_Xnor, Iir_Predefined_Bit_Xnor);
         end if;
         Add_Unary (Name_Not, Iir_Predefined_Bit_Not);
      elsif Decl = Std_Package.Universal_Real_Type then
         List := Create_Iir (Iir_Kind_Interface_List);
         Append_Element (List, Interface);
         Append_Element
           (List,
            Get_Anonymous_Interface (Std_Package.Universal_Integer_Type));
         Add_Operation (Name_Op_Mul, Iir_Predefined_Universal_R_I_Mul,
                        List, Type_Definition);
         Add_Operation (Name_Op_Div, Iir_Predefined_Universal_R_I_Div,
                        List, Type_Definition);

         List := Create_Iir (Iir_Kind_Interface_List);
         Append_Element
           (List,
            Get_Anonymous_Interface (Std_Package.Universal_Integer_Type));
         Append_Element (List, Interface);
         Add_Operation (Name_Op_Mul, Iir_Predefined_Universal_I_R_Mul,
                        List, Type_Definition);
      end if;
   end Create_Implicit_Operations;

   procedure Create_Range_Constraint_For_Enumeration_Type
     (Def : Iir_Enumeration_Type_Definition)
   is
      Range_Expr : Iir_Range_Expression;
      Literal_List: Iir_List;
   begin
      Literal_List := Get_Enumeration_Literal_List (Def);

      --  Create a constraint.
      Range_Expr := Create_Iir (Iir_Kind_Range_Expression);
      Location_Copy (Range_Expr, Def);
      Set_Type (Range_Expr, Def);
      Set_Direction (Range_Expr, Iir_To);
      Set_Left_Limit (Range_Expr, Get_First_Element (Literal_List));
      Set_Right_Limit (Range_Expr, Get_Last_Element (Literal_List));
      Set_Expr_Staticness (Range_Expr, Locally);
      Set_Range_Constraint (Def, Range_Expr);
   end Create_Range_Constraint_For_Enumeration_Type;

   procedure Free_Old_Iir (Node: in Iir)
   is
      N : Iir;
   begin
      N := Node;
      Free_Iir (N);
   end Free_Old_Iir;

   procedure Free_Name (Node : Iir)
   is
      N : Iir;
   begin
      if Node = null then
         return;
      end if;
      N := Node;
      case Get_Kind (N) is
         when Iir_Kind_Simple_Name
           | Iir_Kind_Parenthesis_Name
           | Iir_Kind_Character_Literal
           | Iir_Kind_String_Literal
           | Iir_Kind_Subtype_Definition =>
            Free_Iir (N);
         when Iir_Kind_Selected_Name
           | Iir_Kind_Selected_By_All_Name =>
            Free_Name (Get_Prefix (N));
            Free_Iir (N);
         when Iir_Kind_Library_Declaration
           | Iir_Kind_Package_Declaration
           | Iir_Kind_Design_Unit
           | Iir_Kind_Process_Statement =>
            return;
         when others =>
            Error_Kind ("free_name", Node);
            --Free_Iir (N);
      end case;
   end Free_Name;

   procedure Free_Recursive (Node : Iir; Free_List : Boolean := False)
   is
      N : Iir;
   begin
      if Node = Null_Iir then
         return;
      end if;
      N := Node;
      case Get_Kind (N) is
         when Iir_Kind_Library_Declaration =>
            return;
         when Iir_Kind_Simple_Name
           | Iir_Kind_Parenthesis_Name
           | Iir_Kind_Character_Literal =>
            null;
         when Iir_Kind_Enumeration_Literal =>
            return;
         when Iir_Kind_Selected_Name =>
            Free_Recursive (Get_Prefix (N));
         when Iir_Kind_Constant_Interface_Declaration
           | Iir_Kind_Variable_Interface_Declaration
           | Iir_Kind_Signal_Interface_Declaration =>
            Free_Recursive (Get_Type (N));
            Free_Recursive (Get_Default_Value (N));
         when Iir_Kind_Range_Expression =>
            Free_Recursive (Get_Left_Limit (N));
            Free_Recursive (Get_Right_Limit (N));
         when Iir_Kind_Subtype_Definition =>
            Free_Recursive (Get_Base_Type (N));
         when Iir_Kind_Integer_Literal =>
            null;
         when Iir_Kind_Package_Declaration
           | Iir_Kind_Package_Body
           | Iir_Kind_Entity_Declaration
           | Iir_Kind_Configuration_Declaration =>
            null;
         when Iir_Kind_File_Type_Definition
           | Iir_Kind_Access_Type_Definition
           | Iir_Kind_Array_Type_Definition
           | Iir_Kind_Enumeration_Type_Definition
           | Iir_Kind_Integer_Subtype_Definition
           | Iir_Kind_Enumeration_Subtype_Definition
           | Iir_Kind_Physical_Subtype_Definition =>
            return;
         when Iir_Kind_Architecture_Declaration =>
            Free_Recursive (Get_Entity (N));
         when Iir_Kind_Proxy =>
            null;
         when Iir_Kinds_List =>
            declare
               El : Iir;
            begin
               for I in Natural loop
                  El := Get_Nth_Element (N, I);
                  exit when El = Null_Iir;
                  Free_Recursive (El);
               end loop;
            end;
            if not Free_List then
               return;
            end if;
         when Iir_Kind_Array_Subtype_Definition =>
            Free_Recursive (Get_Index_List (N));
            Free_Recursive (Get_Base_Type (N));
         when Iir_Kind_Entity_Aspect_Entity =>
            Free_Recursive (Get_Entity (N));
            Free_Recursive (Get_Architecture (N));
         when others =>
            Error_Kind ("free_recursive", Node);
      end case;
      Free_Iir (N);
   end Free_Recursive;

   function Get_Predefined_Function_Name (Func : Iir_Predefined_Functions)
                                          return String
   is
   begin
      return Iir_Predefined_Functions'Image (Func);
   end Get_Predefined_Function_Name;

   procedure Clear_Seen_Flag (Top : Iir)
   is
      Callees_List : Iir_Callees_List;
      El: Iir;
   begin
      if Get_Seen_Flag (Top) then
         Set_Seen_Flag (Top, False);
         Callees_List := Get_Callees_List (Top);
         if Callees_List /= null then
            for I in Natural loop
               El := Get_Nth_Element (Callees_List, I);
               exit when El = Null_Iir;
               if Get_Seen_Flag (El) = False then
                  Clear_Seen_Flag (El);
               end if;
            end loop;
         end if;
      end if;
   end Clear_Seen_Flag;

   function Is_Anonymous_Type_Definition (Def : Iir) return Boolean is
   begin
      return Get_Type_Declarator (Def) = Null_Iir;
   end Is_Anonymous_Type_Definition;

   function Is_Unconstrained_Type_Definition (Def : Iir) return Boolean is
   begin
      return Get_Kind (Def) in Iir_Kinds_Unconstrained_Array_Type_Definition;
   end Is_Unconstrained_Type_Definition;

   function Check_Subprograms_Specification (L, R : Iir)
     return Boolean
   is
      List_L, List_R : Iir_List;
      El_L, El_R : Iir;
   begin
      List_L := Get_Interface_Declaration (L);
      List_R := Get_Interface_Declaration (R);
      for I in Natural loop
         El_L := Get_Nth_Element (List_L, I);
         El_R := Get_Nth_Element (List_R, I);
         exit when El_L = Null_Iir and El_R = Null_Iir;
         if El_L = Null_Iir or El_R = Null_Iir then
            return False;
         end if;
         if Get_Base_Type (Get_Type (El_L)) /= Get_Base_Type (Get_Type (El_R))
         then
            return False;
         end if;
      end loop;
      if Get_Kind (L) in Iir_Kinds_Function_Specification then
         if Get_Base_Type (Get_Return_Type (L)) /=
           Get_Base_Type (Get_Return_Type (R))
         then
            return False;
         end if;
      end if;
      return True;
   end Check_Subprograms_Specification;

   -- From a block_specification, returns the block.
   function Get_Block_From_Block_Specification (Block_Spec : Iir)
     return Iir
   is
      Res : Iir;
   begin
      case Get_Kind (Block_Spec) is
         when Iir_Kind_Design_Unit =>
            Res := Get_Library_Unit (Block_Spec);
            if Get_Kind (Res) /= Iir_Kind_Architecture_Declaration then
               raise Internal_Error;
            end if;
            return Res;
         when Iir_Kind_Block_Statement
           | Iir_Kind_Architecture_Declaration
           | Iir_Kind_Generate_Statement =>
            return Block_Spec;
         when Iir_Kind_Indexed_Name
           | Iir_Kind_Selected_Name
           | Iir_Kind_Slice_Name =>
            return Get_Prefix (Block_Spec);
         when others =>
            Error_Kind ("get_block_from_block_specification", Block_Spec);
            return null;
      end case;
   end Get_Block_From_Block_Specification;

   function Get_String_Type_Bound_Type (Sub_Type : Iir) return Iir is
   begin
      if Get_Kind (Sub_Type) /= Iir_Kind_Array_Subtype_Definition then
         Error_Kind ("get_string_type_bound_type", Sub_Type);
      end if;
      return Get_First_Element (Get_Index_List (Sub_Type));
   end Get_String_Type_Bound_Type;

   procedure Get_Low_High_Limit (Arange : Iir_Range_Expression;
                                 Low, High : out Iir)
   is
   begin
      case Get_Direction (Arange) is
         when Iir_To =>
            Low := Get_Left_Limit (Arange);
            High := Get_Right_Limit (Arange);
         when Iir_Downto =>
            High := Get_Left_Limit (Arange);
            Low := Get_Right_Limit (Arange);
      end case;
   end Get_Low_High_Limit;

   function Get_Low_Limit (Arange : Iir_Range_Expression) return Iir is
   begin
      case Get_Direction (Arange) is
         when Iir_To =>
            return Get_Left_Limit (Arange);
         when Iir_Downto =>
            return Get_Right_Limit (Arange);
      end case;
   end Get_Low_Limit;

   function Get_High_Limit (Arange : Iir_Range_Expression) return Iir is
   begin
      case Get_Direction (Arange) is
         when Iir_To =>
            return Get_Right_Limit (Arange);
         when Iir_Downto =>
            return Get_Left_Limit (Arange);
      end case;
   end Get_High_Limit;

end Iirs_Utils;
