with Types; use Types;
with Iirs; use Iirs;

package Iirs_Utils is
   -- Transform the current token into an iir literal.
   -- The current token must be either a character, a string or an identifier.
   function Current_Text return Iir;

   -- Copy the current location into an iir.
   procedure Set_Location (An_Iir: Iir);

   --  Get identifier of NODE as a string.
   function Get_String (Node : Iir) return String;

   --  Find LIT in the list of identifiers or characters LIST.
   --  Return the literal (whose name is LIT) or null_iir if not found.
   function Find_Name_In_List (List: Iir_List; Lit: Name_Id) return Iir;

   --  Return TRUE if EL in an element of list LIST.
   function Is_In_List (List : Iir_List; El : Iir) return Boolean;

   --  Convert an operator node to a name.
   function Get_Operator_Name (Op : Iir) return Name_Id;

   -- Get the longuest static prefix of EXPR.
   -- See LRM 8.1
   function Get_Longuest_Static_Prefix (Expr: Iir) return Iir;

   -- Get the prefix of DECL, ie:
   -- {signal, variable, constant}{interface_declaration, declaration}, or
   -- DECL itself, if it is not an object.
   function Get_Object_Prefix (Decl: Iir) return Iir;

   --  Make TARGETS depends on UNIT.
   --  UNIT must be either a design unit or a entity_aspect_entity.
   procedure Add_Dependence (Target: Iir_Design_Unit; Unit: Iir);

   --  Get Body/spec from a subprogram declaration (body, spec or both).
   function Get_Subprogram_Body (Subprg: Iir) return Iir_List;
   function Get_Subprogram_Spec (Subprg: Iir) return Iir;

   --  Clear configuration field of all component instantiation of LIST.
   procedure Clear_Instantiation_Configuration
     (List : Iir_Concurrent_Statement_List);

   --  Create predefined operations for DECL.
   procedure Create_Implicit_Operations
     (Decl : Iir_Type_Declaration; Is_Std_Standard : Boolean := False);

   --  Free Node and its prefixes, if any.
   procedure Free_Name (Node : Iir);

   --  Free NODE and its sub-nodes.
   procedure Free_Recursive (Node : Iir; Free_List : Boolean := False);

   --  Free NODE.
   procedure Free_Old_Iir (Node: in Iir);

   --  Name of FUNC.
   function Get_Predefined_Function_Name (Func : Iir_Predefined_Functions)
     return String;

   --  Create the range_constraint node for an enumeration type.
   procedure Create_Range_Constraint_For_Enumeration_Type
     (Def : Iir_Enumeration_Type_Definition);

   --  Clear flag of TOP and all of its callees.
   procedure Clear_Seen_Flag (Top : Iir);

   --  Return TRUE iff DEF is an anonymous type (or subtype) definition.
   --  Note: DEF is required to be a type (or subtype) definition.
   --  Note: type (and not subtype) are never anonymous.
   function Is_Anonymous_Type_Definition (Def : Iir) return Boolean;
   pragma Inline (Is_Anonymous_Type_Definition);

   --  Return TRUE iff DEF is an unconstrained type (or subtype) definition.
   function Is_Unconstrained_Type_Definition (Def : Iir) return Boolean;

   --  Return TRUE if L and R have the same signature.
   --  Note: both L and R must be procedure specification
   --   or function specification.
   function Check_Subprograms_Specification (L, R : Iir) return Boolean;

   --  From a block_specification, returns the block.
   --  Roughly speaking, this get prefix of indexed and sliced name.
   function Get_Block_From_Block_Specification (Block_Spec : Iir)
     return Iir;

   --  Return the bound type of a string type, ie the type of the (first)
   --  dimension of a one-dimensional array type.
   function Get_String_Type_Bound_Type (Sub_Type : Iir) return Iir;

   --  Return left or right limit according to the direction.
   procedure Get_Low_High_Limit (Arange : Iir_Range_Expression;
                                 Low, High : out Iir);
   function Get_Low_Limit (Arange : Iir_Range_Expression) return Iir;
   function Get_High_Limit (Arange : Iir_Range_Expression) return Iir;
end Iirs_Utils;

