with Types; use Types;
with Iirs; use Iirs;

package Sem_Expr is
   -- Set semantic to EXPR.
   --  Replace simple_name with the referenced node,
   --  Set type to nodes,
   --  Resolve overloading

   --  Restrictions that apply during semantisation of an expression.
   type Expr_Restriction_Type is private;

   --  Restriction for a value (read).
   Value_Restr: constant Expr_Restriction_Type;

   --  Restriction for an interface (read, deferred constant available).
   Interface_Restr : constant Expr_Restriction_Type;

   --  Restriction for a target (written).
   Target_Restr : constant Expr_Restriction_Type;

   -- If A_TYPE is not null, then EXPR must be of type A_TYPE.
   -- Return null in case of error.
   function Sem_Expression
     (Expr: Iir; A_Type: Iir; Restriction : Expr_Restriction_Type)
     return Iir;

   --  Check EXPR can be read.
   procedure Check_Read (Expr : Iir);

   --  Check EXPR can be updated.
   procedure Check_Update (Expr : Iir);

   --  Level of type compatibility.
   --  For two nodes A and B:
   --  FULL means base type of A and B are the same.
   --  VIA_CONVERSION means that through an implicit conversion
   --    (from universal integer/real) A and B have the same (base) type.
   --  NONE means that base type of A and B are not compatible.
   type Compatibility_Level is (None, Via_Conversion, Full);

   --  Return TRUE iif types of LEFT and RIGHT are compatible.
   function Are_Nodes_Compatible (Left: Iir; Right: Iir)
     return Boolean;

   --  Semantize all arguments of ASSOC_LIST (which can be NULL).
   --  Return TRUE if no error.
   function Sem_Association_List (Assoc_List : Iir_Association_List)
     return Boolean;

   -- Return true iff ASSOC_LIST matches with the subprogram declaration DECL.
   -- If FINISH is true, then assoc_list is reduced.
   -- If CHECK_MATCH is true, then all interface of INTERFACE_LIST must have
   --   an association or a default value.
   --   This may be false for binding indication.
   -- If FINISH is true, then an error message is printed in case of error.
   -- If ALLOW_OPEN is true, then an association may be open even if there is
   --   no default value.
   -- LOC is the association.
   -- Sem_Association_list must have been called before.
   function Sem_Subprogram_Arguments
     (Interface_List : Iir_List;
      Assoc_List: Iir;
      Finish: Boolean;
      Check_Match : Boolean;
      Is_Port : Boolean;
      Loc : Iir)
      return Compatibility_Level;

   --  Semantize a procedure_call or a concurrent_procedure_call_statement.
   procedure Sem_Procedure_Call (Stmt: Iir);

   --  Semantize NAME as long as it consists in declarations.
   function Sem_Names (Name : Iir) return Iir;

   --  Kind of declaration to find.
   --  Decl_entity: an entity declaration (used for binding_indication).
   --  Decl_Any : no checks is performed.

   type Decl_Kind_Type is
     (Decl_Type, Decl_Incomplete_Type, Decl_Object, Decl_Interface,
      Decl_Component, Decl_Physical_Unit, Decl_Label, Decl_Any,
      Decl_Group_Template, Decl_Entity, Decl_Configuration, Decl_Attribute);

   --  Find a uniq declaration for name NAME, which can be a simple_name,
   --  an identifier or a selected_name.
   --  Disp an error message if:
   --   NAME (or any prefix of it) is undefined
   --   NAME is overloaded
   --   NAME does not belong to KIND.
   --  In these case, null is returned.
   --  Otherwise, the declaration is returned, and NAME is freed.
   --  If NAME is a selected_name, dependencies can be added to the current
   --  design unit.
   function Find_Declaration (Name: Iir; Kind: Decl_Kind_Type) return Iir;

   -- Find declarations for a name.
   function Find_Declarations (Name: Iir) return Iir;

   function Sem_Discrete_Range_Expression (Expr: Iir; A_Type: Iir) return Iir;
   function Get_Discrete_Range_Staticness (Expr : Iir) return Iir_Staticness;

   --  Convert a parenthesis_name to a slice_name or an index_name, according
   --  to the suffix expression.
   --  This is used in sem by generates.
   function Sem_Parenthesis_Name (Name : Iir_Parenthesis_Name) return Iir;

   -- Transform LIT into a physical_literal.
   -- LIT can be either a not semantized physical literal or
   --  a simple name that is a physical unit.  In the later case, a physical
   --  literal is created.
   function Sem_Physical_Literal (Lit: Iir) return Iir_Physical_Literal;

   --  CHOICES_LIST is a list of choices (none, expression, range, list or
   --    others).
   --  If IS_SUB_RANGE is true, then SUB_TYPE may not be fully convered,
   --    otherwise, SUB_TYPE must be fully covered.
   --    This is used when the subtype of an aggregate must be determined.
   --  SUB_TYPE is the discrete subtype.
   --  Emit a message if:
   --  * the SUB_TYPE is not fully covered by the choices
   --  * the choices are not mutually exclusif (an element is present twice)
   --  * OTHERS is not the last choice, or is present several times.
   --
   --  If there is at least one named choice, LOW and HIGH are set with the
   --  lowest and highest index.
   --  If LOW and HIGH are set, they are locally static.
   --
   --  Unidimensional strings are not handled here but by
   --  sem_string_choices_range.
   --
   --  TODO:
   --  * be smarter if only positional choices (do not create the list).
   --  * smarter messages.
   procedure Sem_Choices_Range
     (Choice_List : Iir_List;
      Sub_Type : Iir;
      Is_Sub_Range : Boolean;
      Loc : Location_Type;
      Low : out Iir;
      High : out Iir);

   --  Semantize CHOICE_LIST when the choice expression SEL is of a
   --  one-dimensional character array type.
   procedure Sem_String_Choices_Range (Choice_List : Iir_List; Sel : Iir);

private
   --  Restrictions that apply during semantisation of an expression.
   --  To be added:
   --  * the class of the expression (signal, file, variable, expr)
   --  * the minimal staticness required
   --  *
   type Expr_Restriction_Type is record
      --  The mode of the expression.
      --  iir_in_mode means the expression is to be read.
      --    This is the general mode (eg value for an assignment).
      --  iir_out_mode means the expression is to be written.
      --    eg, target of an assignement.
      --  iir_inout_mode means the expression is to be read and written.
      --    eg, actual of an inout procedure parameter.
      Mode : Iir_Mode := Iir_In_Mode;

      --  True if a deferred constant appearing in the scope of its declaration
      --  and before being fully declared may be used.
      --  Related to LRM93 2.6
      Deferred_Const_Allowed: Boolean := False;

      --  True if only local names are allowed.
      --  Set by pure functions (see LRM93 2.2).
      Local_Name_Only : Boolean := False;
   end record;
   pragma Pack (Expr_Restriction_Type);

   Value_Restr: constant Expr_Restriction_Type :=
     (Mode => Iir_In_Mode,
      Deferred_Const_Allowed => False,
      Local_Name_Only => False);

   Interface_Restr : constant Expr_Restriction_Type :=
     (Mode => Iir_In_Mode,
      Deferred_Const_Allowed => True,
      Local_Name_Only => False);

   Target_Restr : constant Expr_Restriction_Type :=
     (Mode => Iir_Out_Mode,
      Deferred_Const_Allowed => False,
      Local_Name_Only => False);
end Sem_Expr;
