with Ada.Text_IO;
with GNAT.Table;
with Types; use Types;
with Name_Table; -- use Name_Table;
with Errorout; use Errorout;
with Iirs_Utils;

package body Sem_Name is
   -- FIXME: names:
   -- scopes => regions ?

   -- An interpretation cell is the element of the simply linked list
   -- of interpratation for an identifier.
   -- DECL is visible declaration;
   -- NEXT is the next element of the list.
   -- Interpretation cells are stored in a stack, Interpretations.
   type Interpretation_Cell is
      record
         Decl: Iir;
         Is_Potential : Boolean;
         Is_Visible : Boolean;
         Next: Name_Interpretation_Type;
      end record;
   pragma Pack (Interpretation_Cell);

   -- To manage the list of interpretation and to add informations to this
   -- list, a stack is used.
   -- Elements of stack can be of kind:
   -- Save_Cell:
   --   the element contains the interpretation INTER for the indentifier ID
   --   for the outer declarative region.
   --   A save cell is always each time a declaration is added to save the
   --   previous interpretation.
   -- Region_Start:
   --   A new declarative region start at interpretation INTER.  Here, INTER
   --   is used as an index in the interpretations stack (table).
   --   ID is used as an index into the unidim_array stack.
   -- Barrier_start, Barrier_end:
   --   All currents interpretations are saved between both INTER, and
   --   are cleared.  This is used to call semantic during another semantic.

   type Scope_Cell_Kind_Type is
      (Save_Cell, Region_Start, Barrier_Start, Barrier_End);
   type Scope_Cell is
      record
         Kind: Scope_Cell_Kind_Type;
         Id: Name_Id;
         Inter: Name_Interpretation_Type;
      end record;

   type Visible_Type_Cell is
      record
         Id: Name_Id;
         Decl: Iir;
      end record;

   package Visible_Types is new GNAT.Table
     (Table_Component_Type => Visible_Type_Cell,
      Table_Index_Type => Visible_Type_Index_Type,
      Table_Low_Bound => No_Visible_Type_Index + 1,
      Table_Initial => 32,
      Table_Increment => 10);

   package Interpretations is new GNAT.Table
     (Table_Component_Type => Interpretation_Cell,
      Table_Index_Type => Name_Interpretation_Type,
      Table_Low_Bound => No_Name_Interpretation + 1,
      Table_Initial => 128,
      Table_Increment => 50);

   package Scopes is new GNAT.Table
     (Table_Component_Type => Scope_Cell,
      Table_Index_Type => Natural,
      Table_Low_Bound => 0,
      Table_Initial => 128,
      Table_Increment => 50);

   -- Index into Interpretations marking the last interpretation of
   -- the previous (immediate) declarative region.
   Current_Scope_Start: Name_Interpretation_Type := No_Name_Interpretation;
   Current_Composite_Types_Start : Visible_Type_Index_Type :=
     No_Visible_Type_Index;

   -- Get and Set the info field of the table table for a
   -- name_interpretation.
   function Get_Interpretation (Id: Name_Id) return Name_Interpretation_Type is
   begin
      return Name_Interpretation_Type (Name_Table.Get_Info (Id));
   end Get_Interpretation;

   procedure Set_Interpretation (Id: Name_Id; Inter: Name_Interpretation_Type)
   is
   begin
      Name_Table.Set_Info (Id, Natural (Inter));
   end Set_Interpretation;

   function Get_Under_Interpretation (Id : Name_Id)
     return Name_Interpretation_Type
   is
      Inter : Name_Interpretation_Type;
   begin
      Inter := Name_Interpretation_Type (Name_Table.Get_Info (Id));

      --  ID has no interpretation.
      --  So, there is no 'under' interpretation (FIXME: prove it).
      if Inter = No_Name_Interpretation then
         return Inter;
      end if;
      for I in reverse Scopes.First .. Scopes.Last loop
         declare
            S : Scope_Cell renames Scopes.Table (I);
         begin
            case S.Kind is
               when Save_Cell =>
                  if S.Id = Id then
                     --  This is the previous one, return it.
                     return S.Inter;
                  end if;
               when Region_Start =>
                  null;
               when Barrier_Start
                 | Barrier_End =>
                  return No_Name_Interpretation;
            end case;
         end;
      end loop;
      return No_Name_Interpretation;
   end Get_Under_Interpretation;

   procedure Check_Interpretations is
      use Ada.Text_IO;
      Inter: Name_Interpretation_Type;
      Last : Name_Interpretation_Type;
      Err : Boolean;
   begin
      Last := Interpretations.Last;
      Err := False;
      for I in 0 .. Name_Table.Last_Name_Id loop
         Inter := Get_Interpretation (I);
         if Inter > Last then
            Put_Line ("bad interpretation for " & Name_Table.Image (I));
            Err := True;
         end if;
      end loop;
      if Err then
         raise Internal_Error;
      end if;
   end Check_Interpretations;

   -- Create a new declarative region.
   -- Simply push a region_start cell and update current_scope_start.
   procedure Open_Declarative_Region is
   begin
      Scopes.Increment_Last;
      Scopes.Table (Scopes.Last) := (Kind => Region_Start,
                                     Inter => Current_Scope_Start,
                                     Id => Name_Id (Visible_Types.Last));
      Current_Scope_Start := Interpretations.Last;
   end Open_Declarative_Region;

   -- Close a declarative region.
   -- Update interpretation of identifiers.
   procedure Close_Declarative_Region is
      Old: Name_Interpretation_Type;
   begin
      loop
         case Scopes.Table (Scopes.Last).Kind is
            when Region_Start =>
               Interpretations.Set_Last (Current_Scope_Start);
               Current_Scope_Start := Scopes.Table (Scopes.Last).Inter;
               Visible_Types.Set_Last
                 (Visible_Type_Index_Type (Scopes.Table (Scopes.Last).Id));
               Scopes.Decrement_Last;
               return;
            when Save_Cell =>
               Old := Get_Interpretation (Scopes.Table (Scopes.Last).Id);
               Set_Interpretation (Scopes.Table (Scopes.Last).Id,
                                   Scopes.Table (Scopes.Last).Inter);
            when Barrier_Start | Barrier_End =>
               raise Internal_Error;
         end case;
         Scopes.Decrement_Last;
      end loop;
   end Close_Declarative_Region;

   procedure Open_Scope_Extension renames Open_Declarative_Region;
   procedure Close_Scope_Extension renames Close_Declarative_Region;

   -- Pointer just past the last barrier_end in the scopes stack.
   Current_Barrier : Integer := 0;

   procedure Push_Interpretations is
   begin
      -- Add a barrier_start.
      -- Save current_scope_start and current_barrier.
      Scopes.Increment_Last;
      Scopes.Table (Scopes.Last) := (Kind => Barrier_Start,
                                     Inter => Current_Scope_Start,
                                     Id => Name_Id (Current_Barrier));

      -- Save all the current name interpretations.
      --  (For each name that have interpretations, there is a save_cell
      --   containing the interpretations for the outer scope).
      for I in Current_Barrier .. Scopes.Last - 1 loop
         if Scopes.Table (I).Kind = Save_Cell then
            Scopes.Increment_Last;
            Scopes.Table (Scopes.Last) :=
              (Kind => Save_Cell,
               Inter => Get_Interpretation (Scopes.Table (I).Id),
               Id => Scopes.Table (I).Id);
            Set_Interpretation (Scopes.Table (I).Id, No_Name_Interpretation);
         end if;
      end loop;

      -- Add a barrier_end.
      -- Save interpretations.last.
      Scopes.Increment_Last;
      Scopes.Table (Scopes.Last) :=
        (Kind => Barrier_End,
         Inter => Interpretations.Last,
         Id => Name_Id (Current_Composite_Types_Start));

      -- Start a completly new scope.
      Current_Scope_Start := Interpretations.Last + 1;

      -- Keep the last barrier.
      Current_Barrier := Scopes.Last + 1;
      Current_Composite_Types_Start := Visible_Types.Last;

      pragma Debug (Name_Table.Assert_No_Infos);
   end Push_Interpretations;

   procedure Pop_Interpretations is
   begin
      -- clear all name interpretations set by the current barrier.
      for I in Current_Barrier .. Scopes.Last loop
         if Scopes.Table (I).Kind = Save_Cell then
            Set_Interpretation (Scopes.Table (I).Id, No_Name_Interpretation);
         end if;
      end loop;
      Scopes.Set_Last (Current_Barrier - 1);
      if Scopes.Table (Scopes.Last).Kind /= Barrier_End then
         raise Internal_Error;
      end if;

      pragma Debug (Name_Table.Assert_No_Infos);

      -- Restore the stack pointer of interpretations.
      Interpretations.Set_Last (Scopes.Table (Scopes.Last).Inter);
      Current_Composite_Types_Start :=
        Visible_Type_Index_Type (Scopes.Table (Scopes.Last).Id);
      Scopes.Decrement_Last;

      -- Restore all name interpretations.
      while Scopes.Table (Scopes.Last).Kind /= Barrier_Start loop
         Set_Interpretation (Scopes.Table (Scopes.Last).Id,
                             Scopes.Table (Scopes.Last).Inter);
         Scopes.Decrement_Last;
      end loop;

      -- Restore current_scope_start and current_barrier.
      Current_Scope_Start := Scopes.Table (Scopes.Last).Inter;
      Current_Barrier := Natural (Scopes.Table (Scopes.Last).Id);

      Scopes.Decrement_Last;
   end Pop_Interpretations;

   -- Return TRUE if INTER was made directly visible via a use clause.
   function Is_Potentially_Visible (Inter: Name_Interpretation_Type)
     return Boolean
   is
   begin
      return Interpretations.Table (Inter).Is_Potential;
   end Is_Potentially_Visible;

   -- Return TRUE iif DECL can be overloaded.
   function Is_Overloadable (Decl: Iir) return Boolean is
   begin
      -- LRM93 10.3:
      -- The overloaded declarations considered in this chapter are those for
      -- subprograms and enumeration literals.
      case Get_Kind (Decl) is
         when Iir_Kind_Enumeration_Literal
           | Iir_Kinds_Function_Specification
           | Iir_Kinds_Procedure_Specification =>
            return True;
         when others =>
            return False;
      end case;
   end Is_Overloadable;

   -- Return true if DECL declare a type that is visible.
   --  This is used to build the list of visible types, ie types that must
   --  be considered for certains expression: access for NULL literals,
   --  arrays and records for aggregates, arrays for string literals.
   function Is_Visible_Type (Decl: Iir) return Boolean is
      Def: Iir;
   begin
      case Get_Kind (Decl) is
         when Iir_Kind_Array_Type_Definition
           | Iir_Kind_Array_Subtype_Definition =>
            raise Internal_Error;
         when Iir_Kind_Type_Declaration =>
            Def := Get_Type (Decl);
         when others =>
            return False;
      end case;
      case Get_Kind (Def) is
         when Iir_Kind_Array_Type_Definition
           | Iir_Kind_Array_Subtype_Definition =>
            return True;
         when Iir_Kind_Record_Type_Definition =>
            return True;
         when Iir_Kind_Access_Type_Definition
           | Iir_Kind_Access_Subtype_Definition =>
            return True;
         when others =>
            return False;
      end case;
   end Is_Visible_Type;

   function Get_Visible_Type (Vt: Visible_Type_Index_Type)
     return Visible_Type_Index_Type
   is
      Pt: Visible_Type_Index_Type := Vt;
   begin
      if True then
         return Pt;
      else
         while Pt > Current_Composite_Types_Start loop
            if Get_Declaration
              (Get_Interpretation (Visible_Types.Table (Pt).Id))
              = Visible_Types.Table (Pt).Decl
            then
               return Pt;
            end if;
            Pt := Pt - 1;
         end loop;
         return No_Visible_Type_Index;
      end if;
   end Get_Visible_Type;

   -- Get the first visible declaration of unidim array.
   function Get_First_Visible_Type return Visible_Type_Index_Type is
   begin
      return Get_Visible_Type (Visible_Types.Last);
   end Get_First_Visible_Type;

   -- Get the next visible declaration of unidim array in the list.
   function Get_Next_Visible_Type (Index: Visible_Type_Index_Type)
     return Visible_Type_Index_Type is
   begin
      return Get_Visible_Type (Index - 1);
   end Get_Next_Visible_Type;

   -- Get the declaration corresponding to an uni_array_visible_type.
   function Get_Visible_Type_Decl (Index : Visible_Type_Index_Type)
     return Iir is
   begin
      return Visible_Types.Table (Index).Decl;
   end Get_Visible_Type_Decl;

   -- Return TRUE if INTER was made direclty visible in the current
   -- declarative region.
   function Is_In_Current_Declarative_Part (Inter: Name_Interpretation_Type)
                                            return Boolean is
   begin
      return Inter > Current_Scope_Start;
   end Is_In_Current_Declarative_Part;

   --  Called when CURR is being declared in the same declarative region as
   --  PREV, using the same identifier.
   --  Return TRUE if this redeclaration is allowed.
   function Redeclaration_Allowed (Prev, Curr : Iir) return Boolean is
   begin
      case Get_Kind (Curr) is
         when Iir_Kinds_Function_Specification
           | Iir_Kinds_Procedure_Specification =>
            if ((Get_Kind (Prev) in Iir_Kinds_User_Function_Specification
              and then
              Get_Kind (Curr) in Iir_Kinds_User_Function_Specification)
              or else
              (Get_Kind (Prev) in Iir_Kinds_User_Procedure_Specification
               and then
              Get_Kind (Curr) in Iir_Kinds_User_Procedure_Specification))
            then
               return
                 not Iirs_Utils.Check_Subprograms_Specification (Prev, Curr);
            else
               return True;
            end if;
         when Iir_Kind_Enumeration_Literal =>
            if Get_Kind (Prev) /= Get_Kind (Curr) then
               return False;
            end if;
            return Get_Type (Prev) /= Get_Type (Curr);
         when others =>
            return False;
      end case;
   end Redeclaration_Allowed;

   procedure Add_In_Visible_List (Id : Name_Id; Decl : Iir) is
   begin
      if Is_Visible_Type (Decl) then
         Visible_Types.Increment_Last;
         Visible_Types.Table (Visible_Types.Last) := (Id => Id, Decl => Decl);
      end if;
   end Add_In_Visible_List;

   -- Add interpretation DECL to the identifier of DECL.
   -- POTENTIALLY is true if the identifier comes from a use clause.
   procedure Add_Name (Decl: Iir; Potentially: Boolean)
   is
      Ident: Name_Id;

      -- Current interpretation of ID.  This is the one before DECL is
      -- added (if so).
      Current_Inter: Name_Interpretation_Type;
      Current_Decl : Iir;

      procedure Save_Current_Interpretation is
      begin
         Scopes.Increment_Last;
         Scopes.Table (Scopes.Last) :=
           (Kind => Save_Cell, Id => Ident, Inter => Current_Inter);
      end Save_Current_Interpretation;

      procedure Add_New_Interpretation is
      begin
         Interpretations.Increment_Last;
         Interpretations.Table (Interpretations.Last) :=
           (Decl => Decl, Next => Current_Inter,
            Is_Potential => Potentially, Is_Visible => False);
         Set_Interpretation (Ident, Interpretations.Last);
         if Get_Kind (Decl) /= Iir_Kind_Design_Unit
           and then Get_Visible_Flag (Decl)
         then
            Add_In_Visible_List (Ident, Decl);
         end if;
      end Add_New_Interpretation;

      Is_Homograph: Boolean;
      use Ada.Text_IO;
   begin
      if Get_Kind (Decl) = Iir_Kind_Design_Unit then
         Ident := Get_Identifier (Get_Library_Unit (Decl));
      else
         Ident := Get_Identifier (Decl);
      end if;

      Current_Inter := Get_Interpretation (Ident);

      if Current_Inter = No_Name_Interpretation then
         -- Very simple: no hide, no overloading.
         Save_Current_Interpretation;
         Add_New_Interpretation;
         return;
      end if;

      if Potentially then
         -- Do not re-add a potential decl
         declare
            Inter: Name_Interpretation_Type := Current_Inter;
         begin
            while Inter /= No_Name_Interpretation loop
               if Get_Declaration (Inter) = Decl then
                  return;
               end if;
               Inter := Get_Next_Interpretation (Inter);
            end loop;
         end;
      end if;

      -- LRM 10.3:
      --  Each of two declarations is said to be a homograph of the other if
      --  both declarations have the same identifier, operator symbol, or
      --  character literal, and overloading is allowed for at most one
      --  of the two.
      --  GHDL: ie, NOT overloading is allowed for both; this is a nand!
      Current_Decl := Get_Declaration (Current_Inter);
      Is_Homograph := not (Is_Overloadable (Current_Decl)
                           and Is_Overloadable (Decl));

      if not Is_Homograph then
         --  Current_Inter and Decl overloads.
         if not Is_In_Current_Declarative_Part (Current_Inter) then
            Save_Current_Interpretation;
         else
            if not Redeclaration_Allowed (Current_Decl, Decl) then
               --  GHDL: may be both potential.
               if Is_Potentially_Visible (Current_Inter) then
                  --  FIXME: Disable Current_Decl
                  if Potentially then
                     return;
                  else
                     --  Add Decl.
                     Current_Inter := No_Name_Interpretation;
                  end if;
               else
                  if not Potentially then
                     Error_Msg_Sem
                       ("redeclaration of " & Disp_Node (Current_Decl)
                        & " defined at " & Disp_Location (Current_Decl), Decl);
                  end if;
                  return;
               end if;
            end if;
         end if;
         Add_New_Interpretation;
         return;
      end if;

      --  LRM93 11.2
      --  If two or more logical names having the same identifier appear in
      --  library clauses in the same context, the second and subsequent
      --  occurences of the logical name have no effect.
      --  The same is true of logical names appearing both in the context
      --  clause of a primary unit and in the context clause of a
      --  corresponding secondary unit.
      --  GHDL:  this test is rather rough, but I don't see any cases where
      --  this test is a false positive (an errouneous true).
      if Decl = Current_Decl
        and then Get_Kind (Decl) = Iir_Kind_Library_Declaration
      then
         return;
      end if;

      -- The current interpretation and the new one are homograph.
      if Is_In_Current_Declarative_Part (Current_Inter) then
         -- They are perhaps visible in the same declarative region.
         if Is_Potentially_Visible (Current_Inter) then
            if Potentially then
               -- LRM93 10.4, item #2
               -- Potentially visible declarations that have the same
               -- designator are not made directly visible unless each of
               -- them is either an enumeration literal specification or
               -- the declaration of a subprogram.
               if Decl = Get_Declaration (Current_Inter) then
                  -- The rule applies only for distinct declaration.
                  -- This handles `use p.all; use P.all;'.
                  return;
               end if;
               -- FIXME: what about a third use clause.
               Save_Current_Interpretation;
               Set_Interpretation (Ident, No_Name_Interpretation);
               return;
            else
               -- LRM93 10.4 item #1
               --  A potentially visible declaration is not made directly
               --  visible if the place considered is within the immediate
               --  scope of a homograph of the declaration.
               -- GHDL: Discard the current potentially visible declaration,
               --  only if it is not an entity declaration, since it is used
               --  to find default binding.
               if Get_Kind (Current_Decl) = Iir_Kind_Design_Unit
                 and then Get_Kind (Get_Library_Unit (Current_Decl))
                 = Iir_Kind_Entity_Declaration
               then
                  Save_Current_Interpretation;
               end if;
               Current_Inter := No_Name_Interpretation;
               Add_New_Interpretation;
               return;
            end if;
         else
            --  There is already a declaration in the current scope.
            if Potentially then
               -- LRM93 10.4 item #1
               -- Discard the new and potentially visible declaration.
               -- However, add the type.
               Add_In_Visible_List (Ident, Decl);
               return;
            else
               -- None of the two declarations are potentially visible, ie
               -- both are visible.
               -- LRM 10.3:
               --  Two declarations that occur immediately within the same
               --  declarative region must not be homographs,
               -- FIXME: unless one of them is the implicit declaration of a
               --  predefined operation.
               declare
                  Prev : Iir;
               begin
                  Prev := Get_Declaration (Current_Inter);
                  Error_Msg_Sem ("identifier `" & Name_Table.Image (Ident)
                                 & "' already used for a declaration",
                                 Decl);
                  Error_Msg_Sem ("previous declaration: " & Disp_Node (Prev),
                                 Prev);
                  return;
               end;
            end if;
         end if;
      end if;

      -- Homograph, not in the same scope.
      -- LRM 10.3:
      -- A declaration is said to be hidden within (part of) an inner
      -- declarative region if the inner region contains an homograph
      -- of this declaration; the outer declaration is the hidden
      -- within the immediate scope of the inner homograph.
      Save_Current_Interpretation;
      Current_Inter := No_Name_Interpretation;  -- Hid.
      Add_New_Interpretation;
   end Add_Name;

   procedure Add_Name (Decl: Iir) is
   begin
      Add_Name (Decl, False);
   end Add_Name;

   procedure Replace_Name (Id: Name_Id; Old : Iir; Decl: Iir)
   is
      Inter : Name_Interpretation_Type;
   begin
      Inter := Get_Interpretation (Id);
      if Get_Declaration (Inter) /= Old then
         raise Internal_Error;
      end if;
      if Get_Next_Interpretation (Inter) /= No_Name_Interpretation then
         raise Internal_Error;
      end if;
      Interpretations.Table (Inter).Decl := Decl;
   end Replace_Name;

   procedure Name_Visible (Ident : Name_Id; Decl : Iir) is
   begin
      if Get_Visible_Flag (Decl) then
         --  A name can be made visible only once.
         raise Internal_Error;
      end if;
      Set_Visible_Flag (Decl, True);
      Add_In_Visible_List (Ident, Decl);
   end Name_Visible;

   procedure Name_Visible (Decl : Iir) is
   begin
      Name_Visible (Get_Identifier (Decl), Decl);
   end Name_Visible;

   function Get_Next_Interpretation (Ni: Name_Interpretation_Type)
                                     return Name_Interpretation_Type is
   begin
      if Ni = No_Name_Interpretation then
         raise Internal_Error;
      end if;
      return Interpretations.Table (Ni).Next;
   end Get_Next_Interpretation;

   function Get_Declaration (Ni: Name_Interpretation_Type)
                             return Iir is
   begin
      if Ni = No_Name_Interpretation then
         raise Internal_Error;
      end if;
      return Interpretations.Table (Ni).Decl;
   end Get_Declaration;

   procedure Add_Implicit_Functions (Decl : Iir_Type_Declaration;
                                     Potentially: Boolean)
   is
      List : Iir_Implicit_Functions_List;
      El : Iir;
   begin
      List := Get_Implicit_Functions_List (Decl);
      for I in Natural loop
         El := Get_Nth_Element (List, I);
         exit when El = null;
         Add_Name (El, Potentially);
      end loop;
   end Add_Implicit_Functions;

   procedure Iterator_Decl (Decl : Iir; Arg : Arg_Type)
   is
      List: Iir;
      El: Iir;
      Sub_Decl: Iir;
   begin
      case Get_Kind (Decl) is
         when Iir_Kinds_Procedure_Specification
           | Iir_Kinds_Function_Specification
           | Iir_Kind_Subtype_Declaration
           | Iir_Kind_Constant_Declaration
           | Iir_Kind_Signal_Declaration
           | Iir_Kind_Variable_Declaration
           | Iir_Kind_File_Declaration
           | Iir_Kind_Component_Declaration
           | Iir_Kind_Attribute_Declaration
           | Iir_Kind_Entity_Declaration
           | Iir_Kind_Package_Declaration =>
            Handle_Decl (Decl, Arg);
         when Iir_Kind_Type_Declaration =>
            -- Handle incomplete type declaration.
            if Get_Kind (Get_Type (Decl)) = Iir_Kind_Incomplete_Type_Definition
            then
               return;
            end if;

            Handle_Decl (Decl, Arg);

            --  Add implicit functions.
            List := Get_Implicit_Functions_List (Decl);
            for I in Natural loop
               El := Get_Nth_Element (List, I);
               exit when El = null;
               Handle_Decl (El, Arg);
            end loop;

            case Get_Kind (Get_Type (Decl)) is
               when Iir_Kind_Enumeration_Type_Definition =>
                  List := Get_Enumeration_Literal_List (Get_Type (Decl));
                  for I in Natural loop
                     El := Get_Nth_Element (List, I);
                     exit when El = null;
                     Handle_Decl (El, Arg);
                  end loop;
               when Iir_Kind_Physical_Subtype_Definition =>
                  Sub_Decl := Get_Base_Type (Get_Type (Decl));
                  El := Get_Primary_Unit (Sub_Decl);
                  Handle_Decl (El, Arg);
                  List := Get_Unit_List (Sub_Decl);
                  for I in Natural loop
                     El := Get_Nth_Element (List, I);
                     exit when El = null;
                     Handle_Decl (El, Arg);
                  end loop;
               when others =>
                  null;
            end case;
         when Iir_Kind_Use_Clause =>
            Handle_Decl (Decl, Arg);
         when Iir_Kind_Library_Clause =>
            El := Get_Library_Declaration (Decl);
            if El /= Null_Iir then
               --  May be empty.
               Handle_Decl (El, Arg);
            end if;
         when Iir_Kind_Design_Unit =>
            --Handle_Decl (Get_Library_Unit (Decl), Arg);
            Handle_Decl (Decl, Arg);

         when Iir_Kind_Procedure_Body
           | Iir_Kind_Function_Body =>
            null;

         when Iir_Kind_Attribute_Specification
           | Iir_Kind_Configuration_Specification
           | Iir_Kind_Disconnection_Specification =>
            null;

         when others =>
            Error_Kind ("iterator_decl", Decl);
      end case;
   end Iterator_Decl;

   --  Make POTENTIALLY (or not) visible DECL.
   procedure Add_Name_Decl (Decl : Iir; Potentially : Boolean) is
   begin
      if Get_Kind (Decl) = Iir_Kind_Use_Clause then
         if not Potentially then
            Use_Names (Get_Selected_Name (Decl));
         end if;
      else
         Add_Name (Decl, Potentially);
      end if;
   end Add_Name_Decl;

   procedure Add_Declaration is
      new Iterator_Decl (Arg_Type => Boolean, Handle_Decl => Add_Name_Decl);

   procedure Iterator_Decl_List (Decl_List : Iir_List; Arg : Arg_Type)
   is
      Decl: Iir;
   begin
      if Decl_List = null then
         return;
      end if;
      for I in Natural loop
         Decl := Get_Nth_Element (Decl_List, I);
         exit when Decl = null;
         Handle_Decl (Decl, Arg);
      end loop;
   end Iterator_Decl_List;

   procedure Add_Declarations is new Iterator_Decl_List
     (Arg_Type => Boolean, Handle_Decl => Add_Declaration);

   procedure Use_Library_All (Design_File_List : Iir_Design_File_List)
   is
      Design_File : Iir_Design_File;
      Design_Unit_List : Iir_Design_Unit_List;
      Design_Unit : Iir_Design_Unit;
   begin
      for I in Natural loop
         Design_File := Get_Nth_Element (Design_File_List, I);
         exit when Design_File = Null_Iir;
         Design_Unit_List := Get_Design_Unit_List (Design_File);
         for J in Natural loop
            Design_Unit := Get_Nth_Element (Design_Unit_List, J);
            exit when Design_Unit = Null_Iir;
            Add_Name (Design_Unit, True);
         end loop;
      end loop;
   end Use_Library_All;

   procedure Use_Names (Names: Iir) is
   begin
      case Get_Kind (Names) is
         when Iir_Kind_Design_File_List =>
            Use_Library_All (Names);
         when Iir_Kind_Declaration_List
           | Iir_Kind_Overload_List =>
            Add_Declarations (Names, True);
         when others =>
            Add_Declaration (Names, True);
      end case;
   end Use_Names;

   procedure Add_Declarations_From_Interface_List (Decl_List: Iir) is
      El: Iir;
   begin
      if Decl_List = null then
         return;
      end if;
      for I in Natural loop
         El := Get_Nth_Element (Decl_List, I);
         exit when El = null;
         Add_Name (El, False);
      end loop;
   end Add_Declarations_From_Interface_List;

   procedure Add_Declarations_From_Concurrent_Statement_List
     (Stmt_List: Iir)
   is
      El: Iir;
      Label: Name_Id;
   begin
      for I in Natural loop
         El := Get_Nth_Element (Stmt_List, I);
         exit when El = null;
         Label := Get_Label (El);
         if Label /= Null_Identifier then
            Add_Name (El, False);
         end if;
      end loop;
   end Add_Declarations_From_Concurrent_Statement_List;

   procedure Add_Context_Clauses (Unit : Iir_Design_Unit) is
   begin
      Add_Declarations (Get_Context_Items (Unit), False);
   end Add_Context_Clauses;

   -- Add declarations from an entity into the current declarative region.
   -- This is needed when an architecture is analysed.
   procedure Add_Entity_Declarations (Entity : Iir_Entity_Declaration)
   is
   begin
      Add_Declaration (Entity, False);
      Add_Declarations_From_Interface_List (Get_Generic_List (Entity));
      Add_Declarations_From_Interface_List (Get_Port_List (Entity));
      Add_Declarations (Get_Declaration_List (Entity), False);
   end Add_Entity_Declarations;

   -- Add declarations from a package into the current declarative region.
   -- This is needed when a package body is analysed.
   procedure Add_Package_Declarations (Decl: Iir_Package_Declaration) is
   begin
      Add_Declarations (Get_Declaration_List (Decl), False);
   end Add_Package_Declarations;

   procedure Add_Component_Declarations (Component: Iir_Component_Declaration)
   is
   begin
      Add_Declarations_From_Interface_List (Get_Generic_List (Component));
      Add_Declarations_From_Interface_List (Get_Port_List (Component));
   end Add_Component_Declarations;

   procedure Extend_Scope_Of_Block_Declarations (Decl : Iir) is
   begin
      case Get_Kind (Decl) is
         when Iir_Kind_Architecture_Declaration
           | Iir_Kind_Block_Statement
           | Iir_Kind_Generate_Statement =>
            Add_Declarations (Get_Declaration_List (Decl), False);
            Add_Declarations_From_Concurrent_Statement_List
              (Get_Concurrent_Statement_List (Decl));
         when others =>
            Error_Kind ("extend_scope_of_block_declarations", Decl);
      end case;
   end Extend_Scope_Of_Block_Declarations;

   procedure Iterator_Init_For_Current_Declarative_Part
     (Iterator : out Declarative_Part_Iterator_Type) is
   begin
      Iterator := (Name => Name_Id'First,
                   Inter => No_Name_Interpretation);
   end Iterator_Init_For_Current_Declarative_Part;

   procedure Iterator_Next (Iterator : in out Declarative_Part_Iterator_Type;
                            Name : out Name_Id;
                            Decl : out Iir) is
   begin
      --  Iterate on all identifiers (until an defined identifier is found)
      loop
         if Iterator.Inter = No_Name_Interpretation then
            Iterator.Inter := Get_Interpretation (Iterator.Name);
         else
            Iterator.Inter := Get_Next_Interpretation (Iterator.Inter);
         end if;
         loop
            --  Iterate on all interpretations of the identifier.
            exit when Iterator.Inter = No_Name_Interpretation;
            if Is_In_Current_Declarative_Part (Iterator.Inter) then
               Name := Iterator.Name;
               Decl := Get_Declaration (Iterator.Inter);
               return;
            else
               Iterator.Inter := Get_Next_Interpretation (Iterator.Inter);
            end if;
         end loop;
         Iterator.Name := Iterator.Name + 1;
         exit when Iterator.Name > Name_Table.Last_Name_Id;
      end loop;
      Decl := null;
      return;
   end Iterator_Next;

   -- Debugging
   procedure Disp_Visible_Types is
      use Ada.Text_IO;
      Index: Visible_Type_Index_Type;
   begin
      Index := Get_First_Visible_Type;
      while Index /= No_Visible_Type_Index loop
         Put_Line (Disp_Node (Get_Visible_Type_Decl (Index)));
         Index := Get_Next_Visible_Type (Index);
      end loop;
   end Disp_Visible_Types;

   procedure Disp_Detailed_Interpretations (Ident : Name_Id)
   is
      use Ada.Text_IO;
      use Name_Table;
      procedure Disp_Type (Str : String; Node : Iir) is
      begin
         Put (Str);
         Put_Line
           (Image (Get_Identifier (Get_Type_Declarator (Node))));
      end Disp_Type;

      Inter: Name_Interpretation_Type;
      Decl : Iir;
   begin
      Put (Name_Table.Image (Ident));
      Put_Line (":");

      Inter := Get_Interpretation (Ident);
      while Inter /= No_Name_Interpretation loop
         Decl := Get_Declaration (Inter);
         Put (' ');
         Put (Iir_Kind'Image (Get_Kind (Decl)));
         Put_Line (", loc: " &
                   Name_Table.Image (Get_File_Name (Decl)) & ':' &
                   Natural'Image (Get_Line_Number (Decl)) & ':' &
                   Natural'Image (Get_Column_Number (Decl)));
         case Get_Kind (Decl) is
            when Iir_Kind_Function_Specification
              | Iir_Kind_Function_Spec_Body
              | Iir_Kind_Implicit_Function_Declaration =>
               Disp_Type ("  return type: ", Get_Return_Type (Decl));
               null;
            when others =>
               null;
         end case;
         Inter := Get_Next_Interpretation (Inter);
      end loop;
   end Disp_Detailed_Interpretations;

   procedure Disp_All_Interpretations
     (Interpretation: Name_Interpretation_Type)
   is
      use Ada.Text_IO;
      Inter: Name_Interpretation_Type;
   begin
      Inter := Interpretation;
      while Inter /= No_Name_Interpretation loop
         Put (Name_Interpretation_Type'Image (Inter));
         Put ('.');
         Put (Iir_Kind'Image (Get_Kind (Get_Declaration (Inter))));
         Inter := Get_Next_Interpretation (Inter);
      end loop;
      New_Line;
   end Disp_All_Interpretations;

   procedure Disp_All_Names is
      use Ada.Text_IO;
      Inter: Name_Interpretation_Type;
   begin
      for I in 0 .. Name_Table.Last_Name_Id loop
         Inter := Get_Interpretation (I);
         if Inter /= No_Name_Interpretation then
            Put (Name_Table.Image (I));
            Put (Name_Id'Image (I));
            Put (':');
            Disp_All_Interpretations (Inter);
         end if;
      end loop;
      Put_Line ("interprations.last = "
                & Name_Interpretation_Type'Image (Interpretations.Last));
      Put_Line ("current_scope_start ="
                & Name_Interpretation_Type'Image (Current_Scope_Start));
   end Disp_All_Names;

   procedure Disp_Scopes is
      use Ada.Text_IO;
   begin
      for I in reverse Scopes.First .. Scopes.Last loop
         declare
            S : Scope_Cell renames Scopes.Table (I);
         begin
            case S.Kind is
               when Save_Cell =>
                  Put ("save_cell: `");
                  Put (Name_Table.Image (S.Id));
                  Put ("', old inter:");
               when Region_Start =>
                  Put ("region_start at");
               when Barrier_Start =>
                  Put ("barrier_start at");
               when Barrier_End =>
                  Put ("barrier_end at");
            end case;
            Put_Line (Name_Interpretation_Type'Image (S.Inter));
         end;
      end loop;
   end Disp_Scopes;
end Sem_Name;
