with Iirs; use Iirs;
with Types; use Types;

package Sem_Name is

   --  The purpose of SEM_NAME package is to handle association between
   --  identifiers and declarations.
   --  Roughly speacking, it implements ch10 of LRM: scope and visibility.
   --
   --  Basic elements are: declarations and declarative region.
   --  Declaration should be understood in the large meaning: any textual
   --   construction declaring an identifier, which can be a label.
   --  A declarative region contains declarations and possibly other
   --   declarative regions.
   --
   --  Rules are scope, visibility and overloading.
   --

   -- Create and close a declarative region.
   -- By closing a declarative region, all declarations made in this region
   --  are discarded.
   procedure Open_Declarative_Region;
   procedure Close_Declarative_Region;

   -- Add interpretation DECL for ID to the current declarative region.
   -- ID is an identifier or a character literal.
   -- Note: ID may be different from get_identifier (DECL), since for example
   --  DECL may be a type definition.
   procedure Add_Name (Decl: Iir);
   pragma Inline (Add_Name);

   -- Add interpretation DECL to the identifier of DECL.
   -- POTENTIALLY is true if the identifier comes from a use clause.
   procedure Add_Name (Decl: Iir; Potentially: Boolean);

   --  Set the visible_flag of DECL to true, and add DECL in the list of
   --  visible types if DECL is a type declaration.
   procedure Name_Visible (Decl : Iir);

   --  Replace the interpretation OLD of ID by DECL.
   --  ID must have a uniq interpretation OLD (ie, it must not be overloaded).
   --  The interpretation must have been done in the current scope.
   --  This is necessary when a concurrent_procedure_call_statement becomes
   --  a component_instantiation_statement.
   procedure Replace_Name (Id: Name_Id; Old : Iir; Decl: Iir);

   -- Interpretation is a simply linked list of what an identifier
   -- means.
   type Name_Interpretation_Type is private;
   No_Name_Interpretation: constant Name_Interpretation_Type;

   -- Get the first interpretation of identifier ID.
   function Get_Interpretation (Id: Name_Id) return Name_Interpretation_Type;
   pragma Inline (Get_Interpretation);

   -- Get the next interpretation from an interpretation.
   function Get_Next_Interpretation (Ni: Name_Interpretation_Type)
                                     return Name_Interpretation_Type;
   pragma Inline (Get_Next_Interpretation);

   -- Get a declaration associated with an interpretation.
   function Get_Declaration (Ni: Name_Interpretation_Type) return Iir;
   pragma Inline (Get_Declaration);

   --  Get the previous interpretation of identifier ID, ie the interpretation
   --  for ID before the current interpretation of ID.
   function Get_Under_Interpretation (Id : Name_Id)
     return Name_Interpretation_Type;

   -- Return TRUE if INTER was made directly visible via a use clause.
   function Is_Potentially_Visible (Inter: Name_Interpretation_Type)
     return Boolean;
   pragma Inline (Is_Potentially_Visible);

   -- Return TRUE if INTER was made direclty visible in the current
   -- declarative part.
   function Is_In_Current_Declarative_Part (Inter: Name_Interpretation_Type)
     return Boolean;
   pragma Inline (Is_In_Current_Declarative_Part);

   -- Push and pop all interpretations.
   -- This can be used to suspend name interpretation, in case of recursive
   -- semantics.
   -- After a push, all names have no_name_interpretation.
   -- Pop restore the previous state.
   procedure Pop_Interpretations;
   procedure Push_Interpretations;

   -- Make visible all implicit functions for type DECL.
   procedure Add_Implicit_Functions (Decl : Iir_Type_Declaration;
                                     Potentially: Boolean);

   -- Execute a use clause on NAMES.
   -- Make potentially directly visible declarations of NAMES.
   procedure Use_Names (Names: Iir);

   --  Add declarations for a context clause into the current declarative
   --  regions.
   procedure Add_Context_Clauses (Unit : Iir_Design_Unit);

   -- Add declarations from an entity into the current declarative region.
   -- This is needed when an architecture is analysed.
   procedure Add_Entity_Declarations (Entity : Iir_Entity_Declaration);

   -- Add declarations from a package into the current declarative region.
   -- This is needed when a package body is analysed.
   -- FIXME:  this must be done as if the declarative region was extended.
   procedure Add_Package_Declarations (Decl: Iir_Package_Declaration);

   --  Add interfaces declaration of a component into the current declarative
   --  region.
   procedure Add_Component_Declarations
     (Component : Iir_Component_Declaration);

   --  Scope extension area contains declarations from another declarative
   --  region.  These area are abstract and only used to be able to add
   --  and remove declarations.
   procedure Open_Scope_Extension;
   procedure Close_Scope_Extension;

   -- Add any declarations that include the end of the declarative part of
   --  the given block BLOCK.  This follow rules of LRM93 10.2
   -- FIXME: BLOCK must be an architecture at first, then blocks declared
   --  inside this architecture, then a block declared inside this block...
   -- This procedure must be called after an Open_Scope_Extension and
   --  declarations added can be removed with Close_Scope_Extension.
   procedure Extend_Scope_Of_Block_Declarations (Decl : Iir);

   -- It is necessary to keep trace of all visible type definition of
   -- arrays, record and access.  This is used by string, bit string, aggregate
   -- and null literal.
   -- This is for the user a simple linked list.

   -- list element type.
   type Visible_Type_Index_Type is private;

   -- End of the list element.
   No_Visible_Type_Index: constant Visible_Type_Index_Type;

   -- Get the first visible type declaration.
   function Get_First_Visible_Type return Visible_Type_Index_Type;
   pragma Inline (Get_First_Visible_Type);

   -- Get the visible type declaration after INDEX.
   function Get_Next_Visible_Type (Index: Visible_Type_Index_Type)
     return Visible_Type_Index_Type;
   pragma Inline (Get_Next_Visible_Type);

   -- Get the declaration corresponding to INDEX.
   function Get_Visible_Type_Decl (Index: Visible_Type_Index_Type)
     return Iir;
   pragma Inline (Get_Visible_Type_Decl);

   --  An iterator provides a way to get all declaration of a declarative
   --  part.
   --  Currently, the only the immediatly enclosing declarative region can
   --  have accessed with the iterator.
   --  An iterator is initialized with iterator_init_for_current_region.
   --  At each call to iterator_next, a name and a decl is returned.
   --  If DECL is null, then there is no more declaration.
   --  There is no restriction on the number of iterator.  However, the
   --  scope must not be modified (ie no calls to add_name) during the use of
   --  an iterator.
   type Declarative_Part_Iterator_Type is private;

   procedure Iterator_Init_For_Current_Declarative_Part
     (Iterator : out Declarative_Part_Iterator_Type);
   procedure Iterator_Next (Iterator : in out Declarative_Part_Iterator_Type;
                            Name : out Name_Id;
                            Decl : out Iir);

   --  Call HANDLE_DECL for each declaration found in DECL.
   --  This will generally call HANDLE_DECL with DECL.
   --  For types, HANDLE_DECL is first called with the type declaration, then
   --  with implicit functions, with element literals for enumeration type,
   --  and units for physical type.
   generic
      type Arg_Type is private;
      with procedure Handle_Decl (Decl : Iir; Arg : Arg_Type);
   procedure Iterator_Decl (Decl : Iir; Arg : Arg_Type);

   --  Call HANDLE_DECL for each declaration found in DECL_LIST.
   --  Generally, HANDLE_DECL must be an ITERATOR_DECL; this is not
   --  automatically done, since the user might be interested in using the
   --  ITERATOR_DECL.
   generic
      type Arg_Type is private;
      with procedure Handle_Decl (Decl : Iir; Arg : Arg_Type);
   procedure Iterator_Decl_List (Decl_List : Iir_List; Arg : Arg_Type);

private
   type Name_Interpretation_Type is range 0 .. (2 ** 30) - 1;
   No_Name_Interpretation: constant Name_Interpretation_Type := 0;

   type Visible_Type_Index_Type is new Natural;
   No_Visible_Type_Index: constant Visible_Type_Index_Type := 0;

   type Declarative_Part_Iterator_Type is
      record
         --  Current identifier.
         Name : Name_Id;

         --  Current interpretation.
         --  Set to no_name_interpretation before the first interpretation.
         Inter : Name_Interpretation_Type;
      end record;
end Sem_Name;
